import uuid
import random
import inspect
import pkg_resources

import wx

from .editors.linked_base import LinkedBase
from .arch import fonts

from sawx import errors
from sawx.utils.sortutil import ranges_to_indexes, collapse_overlapping_ranges

from .utils import searchutil
from .ui.segment_grid import SegmentGridControl
from .viewers.mouse_modes import NormalSelectMode
from .commands import clipboard as clipboard_commands

import logging
log = logging.getLogger(__name__)
caret_log = logging.getLogger("caret")
event_log = logging.getLogger("event")


known_viewers = {}

def get_viewers():
    global known_viewers

    if not known_viewers:
        viewers = {}
        for entry_point in pkg_resources.iter_entry_points('omnivore.viewers'):
            mod = entry_point.load()
            log.debug(f"get_viewers: Found module {entry_point.name}")
            for name, obj in inspect.getmembers(mod):
                if inspect.isclass(obj) and SegmentViewer in obj.__mro__ and obj.name:
                    log.debug(f"get_viewers: Found viewer class {name}")
                    viewers[obj.name] = obj
        known_viewers = viewers
    return known_viewers


def find_viewer_class_by_name(name):
    """Find the editor class given its class name

    Returns the OmnivoreEditor subclass whose `name` class attribute matches
    the given string.
    """
    viewers = get_viewers()
    log.debug(f"finding viewers using {viewers}")
    try:
        return viewers[name]
    except KeyError:
        raise ValueError(f"No viewer named {name}")


class SegmentViewer:
    """Base class for any viewer window that can display (& optionally edit)
    the data in a segment

    Linked base exists for the lifetime of the viewer. If the user wants to
    change the base, a new viewer is created and replaces this viewer.
    """
    ##### class attributes

    name = ""  # slug to uniquely identify viewer class

    ui_name = ""  # text to be used in titles and menus

    viewer_category = "Data"

    control_cls = SegmentGridControl  # override in subclass

    override_table_cls = None  # override grid table

    has_editable_bytes = True  # directly user-editable bytes

    has_bitmap = False

    has_font = False  # uses the bitmapped font to display characters

    has_colors = False  # uses machine colors to display stuff

    has_cpu = False

    has_hex = False

    has_text_font = False  # uses the app-wide text font

    has_width = False  # has a user-specifiable items_per_row

    width_text = "viewer width in bytes"

    has_zoom = False  # has user-specifiable zoom value

    zoom_text = "viewer zoom factor"

    has_caret = True

    valid_mouse_modes = []  # toolbar description

    default_mouse_mode_cls = NormalSelectMode

    popup_menu_desc = [
        "view_width",
        "view_zoom",
        None,
        "copy",
        "cut",
        "paste",
        None,
        ["Copy Special",
            "copy_as_repr",
            "copy_as_c_bytes",
            "copy_disassembly",
            "copy_disassembly_comments",
        ],
        ["Paste Special",
            "paste_comments",
            "paste_disassembly_comments",
        ],
        None,
        "select_all",
        "select_none",
        None,
        ["Mark Selection As",
            "disasm_type",
        ],
        None,
        "segment_from_selection",
        "segment_revert_to_baseline",
        None,
        "segment_add_comment",
        "segment_remove_comment",
        "segment_add_label",
        "segment_remove_label",
        None,
        ["Tile Layout",
            "layout_add_viewer_above",
            "layout_add_viewer_below",
            "layout_add_viewer_left",
            "layout_add_viewer_right",
            None,
            "layout_move_to_top_sidebar",
            "layout_move_to_bottom_sidebar",
            "layout_move_to_left_sidebar",
            "layout_move_to_right_sidebar",
            None,
            "layout_remove_viewer_and_keep_tile",
            "layout_remove_viewer_and_close_tile",
        ],
    ]

    searchers = [  # BaseSearcher classes that are applicable to this viewer
        searchutil.HexSearcher,
        searchutil.CommentSearcher,
    ]

    # List of menu bar titles to be excluded from the menu bar when this viewer
    # is focused
    exclude_from_menubar = ["Jumpman"]

    # List of toolbar items that should only be shown with this viewer
    viewer_extra_toolbar_desc = []

    # Performance modifier to prevent non-essential viewers from getting
    # updated at each frame. Viewers will not be updated frame count reaches a
    # multiple of this value.
    priority_refresh_frame_count = 12

    def __init__(self, control, linked_base):
        self.uuid = str(uuid.uuid4())
        self.control = control
        self.linked_base = linked_base

        self.range_processor = ranges_to_indexes
        self.is_tracing = False
        self.frame_count = 0

    ##### Properties

    @property
    def segment(self):
        return self.linked_base.segment

    @property
    def caret_conversion_segment(self):
        """If the segment used to display the carets and selection is different
        than the segment in the linked base, supply the alternate segment here.

        This can happen when the control uses computed data from the current
        segment, or uses an alternate ordering from the current segment. See
        `HiresPage1Viewer` for an example.
        """
        return self.linked_base.segment

    @property
    def can_copy(self):
        return self.linked_base.can_copy

    @property
    def editor(self):
        return self.linked_base.editor

    @property
    def document(self):
        return self.linked_base.editor.document

    @property
    def frame(self):
        return self.linked_base.editor.frame

    @property
    def preferences(self):
        return self.linked_base.editor.preferences

    @property
    def is_focused_viewer(self):
        return self.linked_base.editor.focused_viewer == self

    ##### Class methods

    @classmethod
    def create_control(cls, parent, linked_base, mdict):
        # if a control isn't based on SegmentGridControl, this is the place for
        # the subclass to return the custom control
        return cls.control_cls(parent, linked_base, mdict, cls)

    @classmethod
    def check_name(cls, name):
        return name == cls.name

    @classmethod
    def is_segment_specific_for_display(cls, segment):
        return False

    @classmethod
    def calc_segment_specific_linked_base(cls, editor):
        # Some viewers need a particular address range to display correctly,
        # e.g. the Apple 2 hires and text modes. If one is needed, create and
        # return it here.
        for s in editor.document.segments:
            log.debug(f"calc_segment_specific_linked_base: checking {s} for specific display in {cls.ui_name}")
            if cls.is_segment_specific_for_display(s):
                linked_base = LinkedBase(editor=editor)
                linked_base.find_segment(s, data_model_changed=False)
                break
        else:
            linked_base = None
        return linked_base

    @classmethod
    def replace_linked_base(cls, linked_base):
        """If the subclass uses a cursor and the cursor isn't indexed the same
        as the default linked_base, this is the opportunity to replace it with
        a different linked base that will disassociate the cursors. This means
        this subclass will not share selection or cursor movement.
        """
        return linked_base

    @classmethod
    def viewer_factory(cls, parent, linked_base, uuid=None, mdict={}):
        linked_base = cls.replace_linked_base(linked_base)
        control = cls.create_control(parent, linked_base, mdict.get('control',{}))
        v = cls(control, linked_base)
        v.restore_session(mdict)

        # if v.machine is None:
        #     print("LOOKING UP MACHINE BY MIME", linked_base.document.mime)
        #     v.machine = Machine.find_machine_by_mime(linked_base.document.mime, default_if_not_matched=True)

        control.segment_viewer = v
        if uuid:
            v.uuid = uuid
        control.uuid = v.uuid
        try:
            control.verify_line_renderer()
        except AttributeError:
            pass
        v.create_post()
        log.debug("create: control=%s, parent=%s uuid=%s" % (control.__class__.__name__, parent.__class__.__name__, v.uuid))
        v.set_event_handlers()
        return v

    @classmethod
    def get_known_subclasses(cls):
        return get_all_subclasses_of(SegmentViewer)

    ##### Cleanup

    def prepare_for_destroy(self):
        log.debug(f"prepare_for_destroy: {self}")
        del self.control.segment_viewer
        try:
            self.control.unmap_events()
        except AttributeError:
            pass
        self.control = None

    def lost_focus(self):
        try:
            self.control.end_editing()
        except AttributeError:
            pass

    ##### Initialization and serialization

    def create_post(self):
        # hook for subclasses to do some extra init
        pass

    def set_event_handlers(self):
        self.document.byte_values_changed_event += self.on_update_table_for_value_change
        self.document.byte_style_changed_event += self.on_update_table_for_style_change
        self.document.structure_changed_event += self.on_recalc_data_model
        self.document.recalc_event += self.on_recalc_view
        self.linked_base.ensure_visible_event += self.on_ensure_visible
        self.linked_base.sync_caret_event += self.on_sync_caret
        self.linked_base.sync_caret_to_index_event += self.on_sync_caret_to_index
        self.document.refresh_event += self.on_refresh_view
        self.linked_base.recalc_event += self.on_recalc_view
        try:
            self.document.priority_level_refresh_event += self.on_priority_level_refresh
            self.document.emulator_breakpoint_event += self.on_emulator_breakpoint
        except AttributeError:
            pass
        else:
            # start the initial frame count on a random value so the frame
            # refresh load can be spread around instead of each with the same
            # frame count being refreshed at the same time.
            self.frame_count = random.randint(0, self.priority_refresh_frame_count)


    def restore_session(self, s):
        log.debug("restore_session: %s" % str(s))
        if 'uuid' in s:
            self.uuid = s['uuid']

        # FIXME: deprecated stuff?
        # if 'machine' in e:
        #     if self.machine is None:
        #         self.machine = Machine()
        #     self.machine.restore_extra_from_dict(e['machine'])
        # elif 'machine mime' in e:
        #     mime = e['machine mime']
        #     if self.machine is None or not mime.startswith(self.machine.mime_prefix):
        #         m = Machine.find_machine_by_mime(mime)
        #         if m is not None:
        #             self.machine = m

        # if 'font' in e or 'font renderer' in e or 'font order' in e:
        #     if 'font renderer' in e or 'font order' in e:
        #         self.set_font(e['font'], e.get('font renderer', None), e.get('font order', None))
        #     else:
        #         self.set_font(e['font'][0], e['font'][1])

        # 'control' is handled during viewer creation process

    def serialize_session(self, s):
        s['name'] = self.name
        s['uuid'] = self.uuid
        s['control'] = {}
        try:
            self.control.serialize_extra_to_dict(s['control'])
        except AttributeError:
            pass

    ##### User interface

    def is_toggle_set(self, toggle_flag):
        return self.linked_base == self.editor.center_base

    ##### SegmentViewer interface

    def update_caption(self):
        self.control.SetName(self.window_title + self.linked_base_segment_identifier)
        self.control.SetLabel(self.ui_name)

    ##### non-wx event handlers

    # Only the refresh event should actually update the screen! All other event
    # handlers should just change the data model without any screen effects.

    def on_ensure_visible(self, evt):
        flags = evt.flags
        if flags.index_visible is not None:
            self.control.keep_index_on_screen(flags.index_visible, flags)

    def on_sync_caret(self, evt):
        flags = evt.flags
        event_log.debug("sync_caret_event: for %s using %s; flags=%s" % (self.control, self.linked_base, str(flags)))
        if self.control == flags.source_control or self.control == flags.advance_caret_position_in_control:
            event_log.debug(f"sync_caret_event: skipping {self.control} because is the source of the carets")
        else:
            event_log.debug("sync_caret_event: syncing %s" % self.control)
            self.sync_caret(flags)

    def sync_caret(self, flags):
        if self.has_caret:
            other = flags.sync_caret_from_control
            if other:
                event_log.debug(f"Converting carets in {self.ui_name} from {other.segment_viewer.ui_name}")
                self.control.caret_handler.convert_carets_from(other.caret_handler)
                if not self.control.keep_current_caret_on_screen(flags):
                    # screen was scrolled, so no need for refresh
                    flags.refreshed_as_side_effect.add(self.control)
            else:
                event_log.debug(f"sync_caret: caret position/selection unchanged")
        else:
            event_log.debug(f"sync_caret: {self.ui_name} refreshed as side effect")
            flags.refreshed_as_side_effect.add(self.control)

    def on_sync_caret_to_index(self, evt):
        flags = evt.flags
        event_log.debug("sync_caret_to_index_event: for %s using %s; flags=%s" % (self.control, self.linked_base, str(flags)))
        if self.control == flags.source_control:
            event_log.debug(f"sync_caret_to_index_event: skipping {self.control} because is the source of the carets")
        else:
            event_log.debug("sync_caret_to_index_event: syncing %s" % self.control)
            self.sync_caret_to_index(flags)

    def sync_caret_to_index(self, flags):
        if self.has_caret:
            self.control.caret_handler.set_caret_from_indexes(flags.caret_index)
            if not self.control.keep_current_caret_on_screen(flags):
                # screen was scrolled, so no need for refresh
                flags.refreshed_as_side_effect.add(self.control)
            else:
                event_log.debug(f"sync_caret: caret position/selection unchanged")

    @property
    def window_title(self):
        return self.ui_name

    @property
    def linked_base_segment_identifier(self):
        """If the viewer is always associated with a particular segment
        override this to return an empty string
        """
        return f" ({self.linked_base.segment.name})"

    def on_recalc_data_model(self, evt):
        """Rebuild the data model after a document formatting (or other
        structural change) or loading a new document.
        """
        self.recalc_data_model()

    def recalc_data_model(self):
        pass

    def on_recalc_view(self, evt):
        event_log.debug(f"on_recalc_view: {self.ui_name}")
        self.recalc_view()

    def recalc_view(self):
        """Rebuild the entire UI after a document formatting (or other
        structural change) or loading a new document.
        """
        self.control.recalc_view()

    def show_caret(self, control, index, bit):
        caret_log.debug("show_caret: %s, index=%d" % (self.ui_name, index))
        self.control.set_caret_index(control, index, bit)

    def on_refresh_view(self, evt):
        event_log.debug(f"on_refresh_view: {self.ui_name} flags={evt.flags}")
        self.refresh_view(evt.flags)

    def refresh_view(self, flags):
        if flags == True:
            log.debug("refresh_event: forcing refresh of %s because no display flags" % self.control)
            self.control.refresh_view()
        elif flags.skip_source_control_refresh and self.control == flags.source_control:
            log.debug("refresh_event: skipping refresh of %s" % self.control)
            # FIXME: the row/col headers aren't refreshed with the call to
            # move_viewport_origin, so force them to be refreshed here. I
            # could just take out the optimization and not skip the source
            # control refresh, but this is not that unpleasant of a hack to
            # save a full refresh of the main grid.
            self.control.refresh_headers()
        elif self.control in flags.refreshed_as_side_effect:
            log.debug("refresh_event: skipping already refreshed %s" % self.control)
        else:
            log.debug("refresh_event: refreshing %s" % self.control)
            self.control.refresh_view()

    def on_update_table_for_value_change(self, evt):
        pass

    def on_update_table_for_style_change(self, evt):
        pass

    def on_priority_level_refresh(self, evt):
        """Refresh based on frame count and priority. If the value passed
        through this event is an integer, all viewers with priority values less
        than the event priority value (i.e. the viewers with a higher priority)
        will be refreshed.
        """
        log.debug("process_priority_level_refresh for %s using %s; flags=%s" % (self.control, self.linked_base, str(evt)))
        count = evt[0]
        self.frame_count += 1
        p = self.priority_refresh_frame_count
        if self.frame_count > p or p < count:
            self.do_priority_level_refresh()
            self.frame_count = 0

    def do_priority_level_refresh(self):
        self.refresh_view(True)

    def on_emulator_breakpoint(self, evt):
        log.debug("process_emulator_breakpoint for %s using %s; flags=%s" % (self.control, self.linked_base, str(evt)))
        self.do_emulator_breakpoint(evt)

    def do_emulator_breakpoint(self, evt):
        self.frame.status_message(f"{self.document.emulator.cycles_since_power_on} cycles")

    #### extra save methods

    def get_extra_segment_savers(self, segment):
        """Hook to provide additional ways to save the data based on this view
        of the data
        """
        return []

    ##### toolbar

    def prune_menubar(self, orig_desc):
        desc = []
        for menu in orig_desc:
            if self.is_menu_valid(menu[0]):
                desc.append(menu)
        return desc

    def is_menu_valid(self, menu):
        return menu not in self.exclude_from_menubar

    def prune_toolbar(self, orig_desc):
        desc = []
        for tool in orig_desc:
            if tool is None or self.is_tool_valid(tool):
                desc.append(tool)
        if self.viewer_extra_toolbar_desc:
            if desc:
                desc.append(None)
            desc.extend(self.viewer_extra_toolbar_desc)
        return desc

    def is_tool_valid(self, menu):
        return True

    def update_toolbar(self):
        self.update_mouse_mode()

    def update_mouse_mode(self, mouse_mode=None):
        self.control.set_mouse_mode(mouse_mode)
        self.control.refresh_view()

    ##### view settings

    def use_default_view_params(self):
        self.control.use_default_view_params()

    def restore_view_params(self, params):
        self.control.restore_view_params(params)

    @property
    def width(self):
        return self.control.items_per_row

    def set_width(self, width):
        self.control.items_per_row = self.validate_width(width)
        wx.CallAfter(self.control.recalc_view)

    def validate_width(self, width):
        return width

    @property
    def zoom(self):
        return self.control.zoom

    def set_zoom(self, zoom):
        self.control.zoom = zoom
        wx.CallAfter(self.control.recalc_view)

    ##### Selections

    def select_all(self):
        self.control.mouse_mode.select_all(self.control.caret_handler)

    def select_none(self):
        self.control.mouse_mode.select_none(self.control.caret_handler)

    def select_invert(self):
        self.control.mouse_mode.select_invert(self.control.caret_handler)

    def highlight_selected_ranges_in_segment(self, selected_ranges, segment):
        # This is default implementation which simply highlights everything
        # between the start/end values of each range. Other selection types
        # (rectangular selection) will need to be defined in the subclass
        segment.set_style_ranges(selected_ranges, selected=True)

    def copy_data_from_selected_ranges(self):
        ranges = self.control.get_selected_ranges()
        log.debug(f"ranges: {ranges}")
        indexes = self.range_processor(ranges)
        log.debug(f"indexes: {list(indexes[0:1000])}")
        data = self.segment[indexes].copy()
        return data

    def copy_data_from_selections(self):
        indexes = self.segment.get_style_indexes(selected=True)
        log.debug(f"indexes: {list(indexes[0:1000])}")
        data = self.segment[indexes].copy()
        return data

    def update_current_selection(self, caret_handler):
        if caret_handler.has_selection:
            message = self.linked_base.set_current_selection(caret_handler)
        else:
            self.linked_base.clear_current_selection()
            message = ""
        self.editor.frame.status_message(message)

    ##### Clipboard & Copy/Paste

    supported_clipboard_data = [
        wx.CustomDataObject("numpy,multiple"),
        wx.CustomDataObject("numpy"),
        wx.CustomDataObject("numpy,columns"),
        wx.TextDataObject(),
        ]

    @property
    def clipboard_data_format(self):
        return "numpy"

    def calc_paste_command(self, clipboard_blob, *args, **kwargs):
        return clipboard_commands.PasteCommand(self.segment, clipboard_blob, *args, **kwargs)

    def calc_delete_command(self):
        pass

    def get_selected_ranges_and_indexes(self):
        return self.control.get_selected_ranges_and_indexes(self.linked_base)

    ##### Status info and text utilities

    def show_status_message(self, msg):
        s = self.get_selected_status_message()
        if s:
            msg = "%s %s" % (msg, s)
        self.frame.status_message(msg)

    def get_address_at_index(self, index):
        return self.segment.get_ui_name_at_index(index)

    def get_label_of_selections(self, carets):
        labels = []
        for start, end in [c.range for c in carets]:
            labels.append("%s-%s" % (self.control.table.get_label_at_index(start), self.get_label_at_index(end - 1)))
        return ", ".join(labels)

    def get_label_of_first_byte(self, ranges):
        labels = []
        for start, end in ranges:
            labels.append(self.control.table.get_label_at_index(start))
        return ", ".join(labels)

    ##### Spring tab (pull out menu) interface

    def __call__(self, parent, task, **kwargs):
        control = self.create(parent, task.active_editor.focused_viewer.linked_base)
        return control

    def activate_spring_tab(self):
        self.recalc_view()


class PlaceholderControl(wx.Window):
    tile_manager_empty_control = True

    def __init__(self, parent, linked_base, mdict, viewer_cls):
        wx.Window.__init__(self, parent)
        self.Bind(wx.EVT_PAINT, self.OnPaint)
        self.Bind(wx.EVT_ERASE_BACKGROUND, self.OnEraseBackground)
        self.Bind(wx.EVT_SIZE, self.OnSize)

    def OnPaint(self, event):
        dc = wx.PaintDC(self)
        self.draw(dc)

    def draw(self, dc):
        size = self.GetClientSize()
        s = "Size: %d x %d"%(size.x, size.y)
        dc.SetFont(wx.NORMAL_FONT)
        w, height = dc.GetTextExtent(s)
        height += 3
        dc.SetBrush(wx.WHITE_BRUSH)
        dc.SetPen(wx.WHITE_PEN)
        dc.DrawRectangle(0, 0, size.x, size.y)
        dc.SetPen(wx.LIGHT_GREY_PEN)
        dc.DrawLine(0, 0, size.x, size.y)
        dc.DrawLine(0, size.y, size.x, 0)
        dc.DrawText(s, (size.x-w)/2, (size.y-height*5)/2)
        pos = self.GetPosition()
        s = "Position: %d, %d" % (pos.x, pos.y)
        w, h = dc.GetTextExtent(s)
        dc.DrawText(s, (size.x-w)/2, ((size.y-(height*5))/2)+(height*3))

    def OnEraseBackground(self, event):
        pass

    def OnSize(self, event):
        size = self.GetClientSize()
        s = "Size: %d x %d"%(size.x, size.y)
        self.SetName(s)
        self.Refresh()

    def recalc_view(self):
        self.refresh_view()

    def refresh_view(self):
        dc = wx.ClientDC(self)
        self.draw(dc)

    def refresh_headers(self):
        pass


class PlaceholderViewer(SegmentViewer):
    name = "placeholder"

    ui_name = "Placeholder"

    control_cls = PlaceholderControl

    has_editable_bytes = False



import os
import tempfile
import inspect
import pkg_resources

import numpy as np

from atrip import find_container

from .debugger import Debugger
from .debugger.dtypes import FRAME_STATUS_DTYPE
from .utils.historyutil import RestartTree
from atrip import disassembler as disasm
from .utils.templateutil import load_memory_map
from . import errors

# INPUT_DTYPE must match history_input_t from libudis.h and input_template_t from libatari800.h
INPUT_DTYPE = np.dtype([
    ("keychar", np.uint8),
    ("keycode", np.uint8),
    ("special_key", np.uint8),
    ("flags", np.uint8),
    ("joystick_triggers", np.uint8),
    ("joysticks", np.uint8, 2),
    ("paddle_triggers", np.uint8),
    ("paddles", np.uint8, 8),
    ("mouse_x", np.uint8),
    ("mouse_y", np.uint8),
    ("mouse_buttons", np.uint8),
    ("unused", np.uint8, 5),
])

import logging
log = logging.getLogger(__name__)

KFEST_HACK = False


# Values must correspond to values in libdebugger.h
FRAME_START = 0
FRAME_FINISHED = 1
FRAME_BREAKPOINT = 2
FRAME_WATCHPOINT = 3


class Emulator(Debugger):
    cpu = "<base>"
    ui_name = "<pretty name>"

    supported_filesystems = []
    supported_binaries = []

    serializable_attributes = ['input_raw', 'output_raw', 'frame_count']
    serializable_computed = {'input_raw', 'output_raw'}

    input_array_dtype = INPUT_DTYPE
    output_array_dtype = None
    width = 320
    height = 192

    low_level_interface = None  # cython module; e.g.: libatari800, lib6502

    history_entry_dtype = disasm.dd.HISTORY_ENTRY_DTYPE

    def __init__(self):
        Debugger.__init__(self)
        self.input_raw = np.zeros([self.input_array_dtype.itemsize], dtype=np.uint8)
        self.input = self.input_raw.view(dtype=self.input_array_dtype)
        self.output_raw = np.zeros([FRAME_STATUS_DTYPE.itemsize + self.output_array_dtype.itemsize], dtype=np.uint8)
        self.status = self.output_raw[0:FRAME_STATUS_DTYPE.itemsize].view(dtype=FRAME_STATUS_DTYPE)
        self.output = self.output_raw[FRAME_STATUS_DTYPE.itemsize:].view(dtype=self.output_array_dtype)
        self.num_stringified_lines = 500
        self.stringified_lines = disasm.StringifiedHistory(self.num_stringified_lines)
        self.bootfile = None
        self.frame_event = []
        self.init_restart_tree()
        self.offsets = None
        self.names = None
        self.save_state_memory_blocks = None
        self.main_memory = None
        self.last_boot_state = None
        self.forced_modifier = None
        self.emulator_started = False
        self.cpu_history = None
        self.labels = None

        self.compute_color_map()
        self.screen_rgb, self.screen_rgba = self.calc_screens()

    @property
    def raw_array(self):
        return self.output_raw

    @property
    def raw_state(self):
        return self.output_raw[self.state_start_offset:]

    @property
    def video_array(self):
        return self.output['video'][0]

    @property
    def audio_array(self):
        return self.output['audio'][0]

    @property
    def state_array(self):
        return self.output['state'][0]

    @property
    def memory_access_array(self):
        return self.status['memory_access'][0]

    @property
    def access_type_array(self):
        return self.status['access_type'][0]

    @property
    def current_frame_number(self):
        return self.status['frame_number'][0]

    @property
    def is_frame_finished(self):
        return self.status['frame_status'][0] == FRAME_FINISHED

    @property
    def current_cycle_in_frame(self):
        return self.status['current_cycle_in_frame'][0]

    @property
    def cycles_user(self):
        return self.status['cycles_user'][0]

    @property
    def cycles_since_power_on(self):
        return self.status['cycles_since_power_on'][0]

    @property
    def instructions_since_power_on(self):
        return self.status['instructions_since_power_on'][0]

    @property
    def scan_lines_since_power_on(self):
        return self.status['scan_lines_since_power_on'][0]

    @property
    def current_scan_line_in_frame(self):
        return self.status['current_scan_line_in_frame'][0]

    @property
    def stack_pointer(self):
        raise NotImplementedError("define stack_pointer property in subclass")

    @stack_pointer.setter
    def stack_pointer(self, value):
        raise NotImplementedError("define stack_pointer property in subclass")

    @property
    def program_counter(self):
        raise NotImplementedError("define program_counter property in subclass")

    @program_counter.setter
    def program_counter(self, value):
        raise NotImplementedError("define program_counter property in subclass")

    @property
    def current_cpu_status(self):
        return "not running"

    @property
    def has_save_points(self):
        return len(self.current_restart) > 0

    @classmethod
    def guess_from_document(cls, document):
        for container in document.collection.containers:
            log.debug(f"guess_from_document: container {container}")
            if container.filesystem.ui_name in cls.supported_filesystems:
                return True
            for segment in container.iter_segments():
                log.debug(f"guess_from_document: segment {segment}")
                if segment.ui_name in cls.supported_binaries:
                    return True
        return False

    ##### Video

    def compute_color_map(self):
        pass

    def calc_screens(self):
        rgb = np.empty((self.height, self.width, 3), np.uint8)
        rgba = np.empty((self.height, self.width, 4), np.uint8)
        return rgb, rgba

    ##### Serialization

    def restore_computed_attributes(self, state):
        Debugger.restore_computed_attributes(self, state)
        self.input_raw[:] = state['input_raw']
        self.restore_state(state['output_raw'])

        # recalculate any internal offsets within output_raw, e.g. atari800
        # needs this because the size of the state save file changes depending
        # on the emulator configuration
        self.low_level_interface.get_current_state(state['output_raw'])
        self.configure_save_state_memory_blocks()

    ##### Initialization

    def configure_emulator(self, emu_args=None, instruction_history_count=200000, *args, **kwargs):
        self.configure_labels()
        self.init_cpu_history(instruction_history_count)
        self.args = self.process_args(emu_args)
        self.low_level_interface.clear_state_arrays(self.input, self.output_raw)
        self.low_level_interface.init_emulator(self.args)
        self.configure_emulator_defaults()
        self.emulator_started = True
        self.configure_io_arrays()

    def configure_labels(self, labels=None):
        if labels is None:
            labels = load_memory_map(self.name)
        self.labels = labels
        self.stringified_lines = disasm.StringifiedHistory(self.num_stringified_lines, self.labels)

    def configure_emulator_defaults(self):
        pass

    def configure_io_arrays(self):
        if not self.emulator_started:
            print(f"configure_io_arrays can't be called before emulator started")
            return
        print(f"input_raw: {self.input_raw.shape}  {self.input_raw}")
        print(f"output_raw: {self.output_raw.shape} {self.output_raw}")
        self.low_level_interface.configure_state_arrays(self.input, self.output_raw)
        self.configure_save_state_memory_blocks()

    def configure_save_state_memory_blocks(self):
        self.parse_state()
        self.generate_save_state_memory_blocks()
        self.cpu_state = self.calc_cpu_data_array()
        self.main_memory = self.calc_main_memory_array()

    def process_args(self, emu_args):
        return emu_args if emu_args else []

    def add_segment_to_memory(self, segment):
        start = segment.origin
        end = (start + len(segment)) & 0xffff
        count = end - start
        log.debug(f"Copying {segment} to memory: {start:#04x}-{end:#04x}")
        self.main_memory[start:end] = segment.data[:count]

    #### cleanup

    def prepare_destroy(self):
        pass

    ##### Machine boot

    def find_default_boot_segment(self, segments):
        for segment in segments:
            if segment.origin > 0:
                return segment
        return None

    def boot_from_segment(self, boot_segment):
        if boot_segment is not None:
            data = boot_segment.data[:]
            origin = boot_segment.origin
            self.boot_from_raw(data, origin)

    def boot_from_raw(self, data, origin):
        if self.bootfile is not None:
            try:
                os.remove(self.bootfile)
                self.bootfile = None
            except:  # MSW raises WindowsError, but that's not defined cross-platform
                log.warning("Unable to remove temporary boot file %s." % self.bootfile)
        fd, self.bootfile = tempfile.mkstemp(".omnivore_boot_segment")
        fh = os.fdopen(fd, "wb")
        fh.write(data)
        fh.close()
        log.debug(f"Created temporary file {self.bootfile} to use as boot disk image")
        self.boot_from_file(self.bootfile)

    def boot_from_file(self, filename):
        container = find_container(filename, True)
        print(f"container: filename={filename} {container}")
        run_addr = None
        for s in container.iter_segments():
            print(f"segment: {s}")
            if s.origin > 0:
                try:
                    run_addr = s.run_address()
                except AttributeError:
                    if run_addr is None:
                        run_addr = s.origin
                self.add_segment_to_memory(s)
        print(f"running at: {hex(run_addr)}")
        self.program_counter = run_addr
        self.last_boot_state = self.calc_current_state()
        self.coldstart()

    def parse_state(self):
        base = np.byte_bounds(self.output_raw)[0]
        self.state_start_offset = np.byte_bounds(self.state_array)[0] - base

        memaccess_offset = np.byte_bounds(self.memory_access_array)[0] - base
        memtype_offset = np.byte_bounds(self.access_type_array)[0] - base
        video_offset = np.byte_bounds(self.video_array)[0] - base
        audio_offset = np.byte_bounds(self.audio_array)[0] - base
        self.save_state_memory_blocks = [
            (memaccess_offset, self.memory_access_array.nbytes, 0, "Memory Access"),
            (memtype_offset, self.access_type_array.nbytes, 0, "Access Type"),
            (video_offset, self.video_array.nbytes, 0, "Video Frame"),
            (audio_offset, self.audio_array.nbytes, 0, "Audio Data"),
        ]

    def generate_save_state_memory_blocks(self):
        pass

    def next_frame(self):
        self.process_key_state()
        if not self.is_frame_finished:
            print(f"next_frame: continuing frame from cycle {self.current_cycle_in_frame} of frame {self.current_frame_number}")
        if KFEST_HACK:
            self.kfest_before_history_count = len(self.cpu_history)
        bpid = self.low_level_interface.next_frame(self.input, self.output_raw, self.debug_cmd, self.cpu_history)
        if self.is_frame_finished:
            self.frame_count += 1
            self.process_frame_events()
            self.save_history()
        self.forced_modifier = None
        return self.get_breakpoint(bpid)

    def process_frame_events(self):
        still_waiting = []
        for count, callback in self.frame_event:
            if self.frame_count >= count:
                log.debug("processing %s", callback)
                callback()
            else:
                still_waiting.append((count, callback))
        self.frame_event = still_waiting

    def end_emulation(self):
        pass

    def debug_video(self):
        """Return text based view of portion of video array, for debugging
        purposes only so it doesn't have to be fast.
        """
        pass

    def debug_state(self):
        """Show CPU status registers
        """
        print(self.current_cpu_status)

    # Emulator user input functions

    def get_special_key_actions(self):
        """If this emulator has any special keys that may be hard to duplicate
        on a modern keyboard, an action name can be added here and it will be
        added to the menubar.
        """
        return []

    def coldstart(self):
        """Simulate an initial power-on startup.
        """
        self.low_level_interface.clear_state_arrays(self.input, self.output_raw)
        self.init_restart_tree()

    def warmstart(self):
        """Simulate a warm start; i.e. pressing the system reset button
        """
        pass

    def keypress(self, ascii_char):
        """Pass an ascii char to the emulator
        """
        self.send_char(ord(ascii_char))

    def joystick(self, stick_num, direction_value, trigger_pressed=False):
        """Pass a joystick/trigger value to the emulator
        """
        pass

    def paddle(self, paddle_num, paddle_percentage):
        """Pass a paddle value to the emulator
        """
        pass

    def process_key_state(self):
        """Read keyboard and compute any values that should be sent to the
        emulator.
        """
        pass

    def process_key_down(self, evt, keycode):
        pass

    def process_key_up(self, evt, keycode):
        pass

    ##### Input routines

    def send_char(self, key_char):
        print(f"sending char: {key_char}")
        self.input['keychar'] = key_char
        self.input['keycode'] = 0
        self.input['special_key'] = 0

    def send_keycode(self, keycode):
        self.input['keychar'] = 0
        self.input['keycode'] = keycode
        self.input['special_key'] = 0

    def send_special_key(self, key_id):
        self.input['keychar'] = 0
        self.input['keycode'] = 0
        self.input['special_key'] = key_id

    def clear_keys(self):
        self.input['keychar'] = 0
        self.input['keycode'] = 0
        self.input['special_key'] = 0

    # Utility functions

    def load_disk(self, drive_num, pathname):
        self.low_level_interface.load_disk(drive_num, pathname)

    def calc_current_state(self):
        return self.output_raw.copy()

    #### history/checkpoints

    def init_restart_tree(self):
        self.frame_count = 0
        self.restart_tree = RestartTree()
        self.current_restart = self.restart_tree.emulator_start

    def get_restart_summary(self):
        return self.restart_tree.get_summary()

    def get_restart_of_frame(self, frame):
        return self.current_restart.get_restart(frame)

    def begin_restart(self):
        cursor = self.current_frame_number
        current = self.current_restart
        if cursor < current.end_frame:
            # we are restarting from someplace earlier in the history, so we
            # need a new restart object
            self.current_restart = self.restart_tree.create_restart(current.restart_number, cursor)

    def save_history(self, force=False):
        # History is saved in a big list, which will waste space for empty
        # entries but makes things extremely easy to manage. Simply delete
        # a history entry by setting it to NONE.
        frame_number = int(self.status['frame_number'][0])
        if force or self.current_restart.is_memorable(frame_number):
            log.debug(f"Saving history at {frame_number}")
            d = self.calc_current_state()
            self.current_restart.save_frame(frame_number, d)
            if KFEST_HACK:
                try:
                    for i in range(self.kfest_before_history_count, len(self.cpu_history)):
                        self.kfest_history_to_frame_number[i] = frame_number
                    self.kfest_frame_number_to_history[frame_number] = (self.kfest_before_history_count, len(self.cpu_history))
                except IndexError:
                    raise
            # self.print_history(frame_number)

    def get_history(self, frame_number):
        frame_number = int(frame_number)
        raw = self.current_restart[frame_number]
        status = raw[0:FRAME_STATUS_DTYPE.itemsize].view(dtype=FRAME_STATUS_DTYPE)
        output = raw[FRAME_STATUS_DTYPE.itemsize:].view(dtype=self.output_array_dtype)
        return status, output

    def restore_state(self, state):
        self.low_level_interface.restore_state(state)
        self.output_raw[:] = state

    def restore_restart(self, restart_number, frame_number):
        print(f"restoring restart {restart_number} from frame number {frame_number}")
        try:
            restart = self.restart_tree[restart_number]
        except IndexError:
            log.error(f"{restart_number} not in history")
        else:
            try:
                d = self.current_restart[frame_number]
            except KeyError:
                log.error(f"{frame_number} not in restart {restart_number}")
            else:
                self.current_restart = restart
                self.restore_state(d)

    def kfest_step_history(self, frame_number, step_to):
        from atrip.disassembler import flags, dd
        from .videolookup import videoarray
        from .utils import apple2util as a2
        try:
            mem_to_screen_xy = self.mem_to_screen_xy
        except AttributeError:
            mem_to_screen_xy = np.zeros(8192, dtype=np.uint16)
            # for addr, y in zip(hgr_offsets, hgr_row_order):
            for y, addr in enumerate(a2.hgr_offsets):
                x = y * 40
                mem_to_screen_xy[addr:addr+40] = np.arange(x, x+40, dtype=np.uint16)
                print(mem_to_screen_xy[addr:addr+40])
            self.mem_to_screen_xy = mem_to_screen_xy
        low, high = self.kfest_frame_number_to_history[frame_number]
        print(f"kfest_step_history: {low}->{high}: {step_to}")
        for row in range(low, step_to):
            h = self.cpu_history[row].view(dtype=dd.HISTORY_6502_DTYPE)
            # print(f"{row}: {h}")
            t = h['disassembler_type']
            if t == flags.DISASM_6502_HISTORY:
                f = h['flag'] & flags.FLAG_RESULT_MASK
                addr = h['target_addr']
                if f == flags.FLAG_MEMORY_ALTER:
                    print("ALTERING", addr, h['after1'])
                    value = h['after1']
                elif f == flags.FLAG_STORE_A_IN_MEMORY:
                    print("ALTERING from A", addr, h['a'])
                    value = h['a']
                elif f == flags.FLAG_STORE_X_IN_MEMORY:
                    print("ALTERING from X", addr, h['x'])
                    value = h['x']
                elif f == flags.FLAG_STORE_Y_IN_MEMORY:
                    print("ALTERING from Y", addr, h['y'])
                    value = h['y']
                else:
                    continue
                self.main_memory[addr] = value
                # self.video_array[step_to] = 1
                if addr >=0x2000 and addr < 0x4000:
                    try:
                        vaddr = mem_to_screen_xy[addr - 0x2000]
                    except IndexError:
                        print("NOT ON SCREEN", addr)
                    else:
                        # bits = a2.byte_to_7_pixels[value]
                        # print("Plotting", bits, "at", first_addr)
                        # self.video_array[first_addr:first_addr+7] = bits
                        self.video_array[vaddr] = value
                    # self.video_array[addr - 0x2000] = value
                    # try:
                    #     dest = videoarray[addr]
                    #     self.video_array[dest] = value
                    #     print(addr, dest, value)
                    # except KeyError:
                    #     print("BAD!!!!", addr, dest)
                self.memory_access_array[addr] = 0xff

    def restore_restart_plus(self, restart_number, frame_number, num_instructions):
        self.restore_restart(restart_number, frame_number)
        self.step_into(num_instructions)
        bpid = self.low_level_interface.next_frame(self.input, self.output_raw, self.debug_cmd, self.cpu_history)
        if self.is_frame_finished:
            self.frame_count += 1
            self.process_frame_events()
            self.save_history()
        self.forced_modifier = None
        return self.get_breakpoint(bpid)

    def restore_history(self, frame_number):
        print(("restoring state from frame %d" % frame_number))
        frame_number = int(frame_number)
        if frame_number < 0:
            return
        try:
            d = self.current_restart[frame_number]
        except KeyError:
            log.error(f"{frame_number} not in history")
            pass
        else:
            self.restore_state(d)
            # self.current_restart[(frame_number + 1):] = []  # remove frames newer than this
            # print(("  %d items remain in history" % len(self.current_restart)))
            # self.frame_event = []

    def print_history(self, frame_number):
        d = self.current_restart[frame_number]
        status = d[:FRAME_STATUS_DTYPE.itemsize].view(dtype=FRAME_STATUS_DTYPE)
        output = d[FRAME_STATUS_DTYPE.itemsize:].view(dtype=self.output_array_dtype)
        print("history[%d] of %d: %d %s" % (status['frame_number'], len(self.current_restart), len(d), output['state'][0][0:8]))

    def get_previous_history(self, frame_cursor):
        return self.current_restart.get_previous_frame(frame_cursor)

    def get_next_history(self, frame_cursor):
        return self.current_restart.get_next_frame(frame_cursor)

    # graphics

    def get_color_indexed_screen(self, frame_number=-1):
        """Return color indexed screen in whatever native format this
        emulator supports
        """
        pass

    def get_frame_rgb(self, frame_number=-1):
        """Return RGB image of the current screen
        """

    def get_frame_rgba(self, frame_number=-1):
        """Return RGBA image of the current screen
        """

    def get_frame_rgba_opengl(self, frame_number=-1):
        """Return RGBA image of the current screen, suitable for use with
        OpenGL (flipped vertically)
        """

    # CPU history

    def init_cpu_history(self, num_entries):
        self.cpu_history = disasm.HistoryStorage(num_entries)
        if KFEST_HACK:
            self.kfest_history_to_frame_number = [0]*num_entries
            self.kfest_frame_number_to_history = {}

    def cpu_history_show_range(self, from_index, details=False):
        self.cpu_history.debug_range(from_index)

    def cpu_history_show_next_instruction(self):
        self.low_level_interface.show_next_instruction(self.cpu_history)

    def calc_stringified_history(self, start_index, count):
        """Returns an list of indexes into the entries array for the
        range of history requested.

        Because it's a circular buffer, the wraparound case is unable to be
        handled by numpy, so instead is used to map the row number in the
        display window (which ranges from 0 -> count) to the history entry
        starting at first_entry_index + start_index for count entries.
        """
        return self.cpu_history.stringify_to(self.stringified_lines, start_index, count)

    @property
    def num_cpu_history_entries(self):
        return len(self.cpu_history)


_emulators = None

default_emulator = None

default_emulator_precidence = ["atari800xl", "atari800", "crabapple", "6502"]

def find_first_emulator(precidence, emulators=None):
    if emulators is None:
        emulators = find_emulators()
    for name in precidence:
        for e in emulators:
            if e.name == name:
                return e

def _find_emulators():
    global default_emulator

    emulators = []
    for entry_point in pkg_resources.iter_entry_points('omnivore.emulators'):
        mod = entry_point.load()
        log.debug(f"find_emulators: Found module {entry_point.name}={mod.__name__}")
        for name, obj in inspect.getmembers(mod):
            if inspect.isclass(obj) and Emulator in obj.__mro__[1:]:
                if obj not in emulators:
                    log.debug(f"find_emulators:   found emulator class {name}")
                    emulators.append(obj)
    default_emulator = find_first_emulator(default_emulator_precidence, emulators)
    log.debug(f"find_emulators: Found default emulator {default_emulator}")
    emulators.sort(key=lambda e:e.ui_name)
    return emulators

def find_emulators():
    global _emulators

    if _emulators is None:
        _emulators = _find_emulators()
    return _emulators

def find_emulator(emulator_name):
    for e in find_emulators():
        log.debug(f"find_emulator: looking for {emulator_name}, checking {e.name}")
        if e.name == emulator_name or e == emulator_name:
            return e
    raise errors.UnknownEmulatorError("Unknown emulator '%s'" % emulator_name)

def guess_emulator(document):
    for e in find_emulators():
        log.debug(f"trying emulator {e.ui_name} for {document}")
        if e.guess_from_document(document):
            return e
    raise errors.UnknownEmulatorError(f"No emulator for {document.mime}")



import os
import sys

import wx
import numpy as np

from sawx.utils.nputil import intscale
from sawx.ui import compactgrid as cg

from ..ui.segment_grid import SegmentGridControl, SegmentTable
from ..arch.fonts import AnticFont, valid_fonts
from atrip.char_mapping import valid_font_mappings
from ..arch.font_renderers import valid_font_renderers
from .antic import AnticColorViewer

import logging
log = logging.getLogger(__name__)


class AnticCharImageCache(cg.DrawTextImageCache):
    def draw_item(self, parent, dc, rect, data, style):
        start = 0
        end = len(data)
        nr = 1
        data = data.reshape((nr, -1))
        style = style.reshape((nr, -1))
        v = parent.segment_viewer
        array = parent.font_renderer.get_image(v, v.antic_font, data, style, start, end, end, nr, start, end)
        width = array.shape[1]
        height = array.shape[0]
        if width > 0 and height > 0:
            array = intscale(array, parent.zoom_h, parent.zoom_w)
            image = wx.Image(array.shape[1], array.shape[0])
            image.SetData(array.tobytes())
            bmp = wx.Bitmap(image)
            dc.DrawBitmap(bmp, rect.x, rect.y)


class AnticCharRenderer(cg.TableLineRenderer):
    default_image_cache = AnticCharImageCache

    def __init__(self, parent, image_cache=None):
        image_cache = AnticCharImageCache()
        w = parent.font_renderer.char_bit_width * parent.zoom_w
        h = parent.font_renderer.char_bit_height * parent.zoom_h
        cg.LineRenderer.__init__(self, parent, w, h, parent.items_per_row, image_cache)

    # BaseLineRenderer interface

    def draw_line(self, parent, dc, line_num, col, index, last_index):
        t = parent.table
        rect = self.col_to_rect(line_num, col)
        data = t.data[index:last_index]
        style = t.style[index:last_index]
        self.image_cache.draw_item(parent, dc, rect, data, style)

    # fast BaseLineRenderer interface drawing entire grid at once

    def draw_grid(self, grid_control, dc, first_row, visible_rows, first_cell, visible_cells):
        t = grid_control.table
        first_col = self.cell_to_col(first_row, first_cell)
        last_cell = min(first_cell + visible_cells, self.num_cells)
        last_row = min(first_row + visible_rows, t.num_rows)
        last_col = self.cell_to_col(last_row, last_cell - 1) + 1
        log.debug("draw_grid: rows:%d,%d (vis %d, num %d) cols:%d,%d" % (first_row, last_row, visible_rows, t.num_rows, first_col, last_col))

        ul_rect = self.col_to_rect(first_row, first_col)
        lr_rect = self.col_to_rect(last_row - 1, last_col - 1)
        frame_rect = wx.Rect(ul_rect.x, ul_rect.y, lr_rect.x - ul_rect.x + lr_rect.width, lr_rect.y - ul_rect.y + lr_rect.height)
        dc.SetClippingRegion(frame_rect)

        # First and last rows may not span entire width. Process those
        # separately
        #
        # First row may not have bytes at the beginning of the row if the start
        # offset is not zero
        if first_row == 0:
            try:
                col, index, last_index = self.calc_column_range(grid_control, first_row, first_col, last_col)
            except IndexError:
                pass  # skip lines with no visible cells
            else:
                self.draw_line(grid_control, dc, first_row, col, index, last_index)
            first_row += 1
            if first_row == last_row:
                return
            frame_rect.y += ul_rect.height

        # Last row may not have bytes at the end of the row
        if last_row == t.num_rows:
            try:
                col, index, last_index = self.calc_column_range(grid_control, last_row - 1, first_col, last_col)
            except IndexError:
                pass  # skip lines with no visible cells
            else:
                self.draw_line(grid_control, dc, last_row - 1, col, index, last_index)
            last_row -= 1

        bytes_per_row = t.items_per_row
        nr = last_row - first_row
        nc = last_col - first_col
        first_index = (first_row * bytes_per_row) - t.start_offset
        last_index = (last_row * bytes_per_row) - t.start_offset
        t = t
        if last_index > len(t.data):
            last_index = len(t.data)
            data = np.zeros((nr * bytes_per_row), dtype=np.uint8)
            data[0:last_index - first_index] = t.data[first_index:last_index]
            style = np.zeros((nr * bytes_per_row), dtype=np.uint8)
            style[0:last_index - first_index] = t.style[first_index:last_index]
        else:
            data = t.data[first_index:last_index]
            style = t.style[first_index:last_index]
        data = data.reshape((nr, -1))
        style = style.reshape((nr, -1))

        # get_image(cls, machine, antic_font, byte_values, style, start_byte, end_byte, bytes_per_row, nr, start_col, visible_cols):

        array = grid_control.font_renderer.get_image(grid_control.segment_viewer, grid_control.segment_viewer.antic_font, data, style, first_index, last_index, bytes_per_row, nr, first_col, nc)
        width = array.shape[1]
        height = array.shape[0]
        if width > 0 and height > 0:
            array = intscale(array, grid_control.zoom_h, grid_control.zoom_w)
            #print("bitmap: %d,%d,3 after scaling: %s" % (height, width, str(array.shape)))
            image = wx.Image(array.shape[1], array.shape[0])
            image.SetData(array.tobytes())
            bmp = wx.Bitmap(image)
            dc.DrawBitmap(bmp, frame_rect.x, frame_rect.y)


class CharGridControl(SegmentGridControl):
    initial_zoom = 2

    def set_viewer_defaults(self):
        self.items_per_row = self.view_params.map_width
        self.zoom = 2
        self.inverse = 0

    def calc_line_renderer(self):
        if hasattr(self, 'segment_viewer'):
            return AnticCharRenderer(self)
        return SegmentGridControl.calc_line_renderer(self)

    def verify_line_renderer(self):
        self.recalc_line_renderer()

    @property
    def font_renderer(self):
        return self.segment_viewer.font_renderer

    @property
    def zoom_w(self):
        return self.zoom * self.font_renderer.scale_width

    @property
    def zoom_h(self):
        return self.zoom * self.font_renderer.scale_height

    def recalc_view(self):
        self.table = SegmentTable(self.segment_viewer.linked_base, self.items_per_row)
        self.line_renderer = self.calc_line_renderer()
        SegmentGridControl.recalc_view(self)

    def start_editing(self, evt):
        c = evt.GetKeyCode()
        print(f"start_editing: {c} for {self}")
        if c != wx.WXK_NONE:
            c = self.segment_viewer.font_mapping.convert_byte_mapping(c)
            self.process_edit(c| self.inverse)


class CharViewer(AnticColorViewer):
    name = "char"

    ui_name = "Character"

    control_cls = CharGridControl

    has_font = True

    has_width = True

    width_text = "width in characters"

    has_zoom = True

    zoom_text = "character zoom factor"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._antic_font = None
        self._antic_font_uuid = "e46c1a08-b718-de27-3303-6e2701f0b0b3"  # atari800 default font
        self._antic_font_data = None
        self._font_renderer_name = "Antic 2 (Gr 0)"
        self._font_renderer = None
        self._font_mapping_name = "ASCII Order"
        self._font_mapping = None

    @property
    def window_title(self):
        return self.font_renderer_name + ", " + self.font_mapping_name + ", " + self.color_standard_name

    #### Fonts

    @property
    def font_mapping_name(self):
        return self._font_mapping_name

    @font_mapping_name.setter
    def font_mapping_name(self, value):
        self._font_mapping_name = value
        self._font_mapping = None
        self.graphics_properties_changed()

    @property
    def font_mapping(self):
        if self._font_mapping is None:
            self._font_mapping = valid_font_mappings[self._font_mapping_name]
        return self._font_mapping

    @property
    def font_renderer_name(self):
        return self._font_renderer_name

    @font_renderer_name.setter
    def font_renderer_name(self, value):
        self._font_renderer_name = value
        self._font_renderer = None
        self.graphics_properties_changed()

    @property
    def font_renderer(self):
        if self._font_renderer is None:
            self._font_renderer = valid_font_renderers[self._font_renderer_name]
        return self._font_renderer

    @property
    def antic_font(self):
        if self._antic_font is None:
            self._antic_font = AnticFont(self, self.antic_font_data, self.font_renderer, self.antic_color_registers[4:9])
        return self._antic_font

    @property
    def antic_font_uuid(self):
        return self._antic_font_uuid

    @antic_font_uuid.setter
    def antic_font_uuid(self, value):
        self._antic_font_uuid = value
        self._antic_font_data = None
        self._antic_font = None
        self.graphics_properties_changed()

    @property
    def antic_font_data(self):
        if self._antic_font_data is None:
            self._antic_font_data = valid_fonts[self._antic_font_uuid]
        return self._antic_font_data

    @antic_font_data.setter
    def antic_font_data(self, value):
        self._antic_font_uuid = value["uuid"]
        self._antic_font_data = value
        self._antic_font = None
        self.graphics_properties_changed()

    def colors_changed(self):
        """Hook for subclasses that need to invalidate stuff when colors change
        """
        self._antic_font = None



from builtins import object
from mock import *
from atrip import utils


class TestTextToInt:
    def setup(self):
        pass

    @pytest.mark.parametrize("text,expected,default_base", [
        ("12", 0x12, "hex"),
        ("$1234", 0x1234, "hex"),
        ("0xffff", 0xffff, "hex"),
        ("#12", 12, "hex"),
        ("%11001010", 202, "hex"),

        ("12", 12, "dec"),
        ("$1234", 0x1234, "dec"),
        ("0xffff", 0xffff, "dec"),
        ("#12", 12, "dec"),
        ("%11001010", 202, "dec"),
        ])
    def test_text_to_int(self, text, expected, default_base):
        assert expected == utils.text_to_int(text, default_base)

class TestRanges:
    def setup(self):
        pass

    def test_normal(self):
        values = [0, 1, 2, 3, 4, 5, 7, 10, 11, 99, 500, 501, 502, 892]
        ranges = utils.collapse_to_ranges(values)
        print(ranges)
        assert ranges == [[0, 6], [7, 8], [10, 12], [99, 100], [500, 503], [892, 893]]


    def test_compact(self):
        values = [0, 1, 2, 3, 4, 5, 7, 10, 11, 99, 500, 501, 502, 892]
        ranges = utils.collapse_to_ranges(values, True)
        print(ranges)
        assert ranges == [[0, 6], 7, [10, 12], 99, [500, 503], 892]


if __name__ == "__main__":
    t = TestRanges()
    t.setup()
    t.test_normal()
    t.test_compact()



import os

import numpy as np

import logging
log = logging.getLogger(__name__)


# size_codes and display_codes are combined into a single byte xxxxxxyy
# display_code = x, bytes_per_item = size_code + 1 (1 - 4 bytes per item)

size_codes = {
    'b': 0, # "byte",
    'w': 1, # "word",
    'l': 3, # "long",
    }

display_codes = {
    'h': 0x00, # "hex",
    'he': 0x00,
    'hex': 0x00,
    'c': 0x10, # "char", "ascii",
    'ch': 0x10,
    'cha': 0x10,
    'char': 0x10,
    'as': 0x10,
    'asc': 0x10,
    'asci': 0x10,
    'ascii': 0x10,
    'a': 0x20, # "atascii",
    'at': 0x20,
    'ata': 0x20,
    'atas': 0x20,
    'atasc': 0x20,
    'atasci': 0x20,
    'atascii': 0x20,
    'd': 0x30, # "decimal",
    'de': 0x30,
    'dec': 0x30,
    'deci': 0x30,
    'decim': 0x30,
    'decima': 0x30,
    'decimal': 0x30,
    'b': 0x40, # "binary",
    'bi': 0x40,
    'bin': 0x40,
    'bina': 0x40,
    'binar': 0x40,
    'binary': 0x40,
}


label_description_record_type = np.dtype([  # must match label_description_t
    ('text_length', 'u1'),
    ('num_bytes', 'u1'),
    ('item_count', 'u1'),
    ('type_code', 'u1'),  # size_code | display_code as above
    ('label', 'u1', 12),
    ])

memory_map_record_type = np.dtype([  # must match label_storage_t
    ('flags', 'u2'),
    ('first_addr', 'u2'),
    ('last_addr', 'u2'),
    ('num_labels', 'u2'),
    ('index', 'u2', 256*256),
    ('labels', label_description_record_type, 1024),
    ])


class MemoryMap:
    @classmethod
    def from_file(cls, filename):
        text = open(filename).read()
        return cls.from_text(os.path.basename(filename), text)

    @classmethod
    def from_text(cls, name, text, default_display_key="hex"):
        """Parse lines of text in the format:

            0011 BRKKEY
            0012 RTCLOK        3b
            02a3 TABMAP        15b    binary
            030b DAUX2
            030c TIMER1        w
            0340 IOCB0         16b
            d000 M0PF/HPOSP0

        with 2 required fields and 1 or 2 optional fields, all separated by
        whitespace. The first field is the address, the 2nd is the label, the
        3rd is the data type, and the 4th is the display type. and the
        (optional) 3rd is the size and type of the memory address.

        If the 2nd field contains a "/" character, that will denote a memory
        address that has different uses when read or written to. The read
        address is before the "/", the write address after. Labels may contain
        alphanumeric characters or underscores.

        The optional 3rd field contains the data type; if it is missing, the
        data type is assumed to be a single byte. The data type description is
        one of "b", "w", or "l" (case insensitive), meaning a byte, word (2
        byte), or long-word (4 byte) field. It can be prefixed with a number
        indicating an array of that data type. E.g. "4w" is an array of 4 two-
        byte words.

        The optional 4th field contains a keyword for how the data should be
        presented: "hex", "char", "atascii", "binary", "decimal" or
        abbreviations of those keywords. If missing, the data will be displayed
        in the default type.
        """
        rw = cls(name)
        r = cls(name + " (R)")
        w = cls(name + " (W)")
        try:
            text = text.decode('utf-8')
        except AttributeError as e:
            # already a string!
            pass
        line_num = 0
        for line in text.splitlines():
            line = line.strip()
            line_num += 1
            if not line or line.startswith("#") or line.startswith(";"):
                continue
            line = line.split("#")[0]
            tokens = line.split()
            count = len(tokens)
            if count < 2:
                log.warning(f"Missing label on line {line_num}")
                continue
            try:
                addr = int(tokens[0], 16)
            except ValueError:
                log.warning(f"Invalid hex digits on line {line_num}: '{tokens[0]}'")
                continue
            if "/" in tokens[1]:
                rw_name = None
                r_name, w_name = tokens[1].split("/")
            else:
                rw_name = tokens[1]
                r_name = None
                w_name = None
            try:
                type_key = tokens[2]
            except IndexError:
                type_key = 'b'
            try:
                display_key = tokens[3]
            except IndexError:
                display_key = default_display_key
            try:
                if rw_name is not None:
                    rw.add(addr, rw_name, type_key, display_key)
                if r_name is not None:
                    r.add(addr, r_name, type_key, display_key)
                if w_name is not None:
                    w.add(addr, w_name, type_key, display_key)
            except RuntimeError:
                log.warning("Invalid type code on line {line_num}: `'{type_code}'")
        return rw, r, w

    @classmethod
    def from_list(cls, name, entries, default_display_key="hex"):
        """Parse list of tuples, each tuple in the form of:

            (addr, name[, type_key [, display_key]])

        The addr field must be an integer, name must be a string without a
        slash, and the type and display keys are strings as in the field
        descriptions of `from_text`.
        """
        rw = cls(name)
        line_num = 0
        for tokens in entries:
            count = len(tokens)
            if count < 2:
                log.warning(f"Missing label on line {line_num}")
                continue
            addr = tokens[0]
            rw_name = tokens[1]
            try:
                type_key = tokens[2]
            except IndexError:
                type_key = 'b'
            try:
                display_key = tokens[3]
            except IndexError:
                display_key = default_display_key
            try:
                rw.add(addr, rw_name, type_key, display_key)
            except RuntimeError:
                log.warning("Invalid type code on line {line_num}: `'{type_code}'")
        return rw

    def __init__(self, name):
        self.ui_name = name
        self.labels_raw = np.zeros([memory_map_record_type.itemsize], dtype=np.uint8)
        self.labels = self.labels_raw.view(dtype=memory_map_record_type)
        self.next_index = 1  # index 0 is unused, indicates unused address

    def __str__(self):
        lines = []
        valid = np.where(self.labels['index'][0] > 0)[0]
        for addr in valid:
            index = self.labels['index'][0][addr]
            r = self.labels['labels'][0][index]
            label_size = r['text_length']
            label = "".join(map(chr, r['label'][0:label_size]))
            num_bytes = r['num_bytes']
            item_count = r['item_count']
            type_code = r['type_code']
            lines.append(f"{addr:4x}: {label:8} {num_bytes} {item_count} {type_code}")
        return "\n".join(lines)

    def __contains__(self, addr):
        return self.labels['index'][0][addr] > 0

    def __getitem__(self, addr):
        if isinstance(addr, int):
            index = self.labels['index'][0][addr]
            if index == 0:
                raise KeyError(f"No label at {addr:x}")
            r = self.labels['labels'][0][index]
            label_size = r['text_length']
            label = "".join(map(chr, r['label'][0:label_size]))
            num_bytes = r['num_bytes']
            item_count = r['item_count']
            type_code = r['type_code']
            return (label, num_bytes, item_count, type_code)
        elif isinstance(addr, slice):
            raise TypeError(f"slicing not yet supported")
        else:
            raise TypeError(f"index must be int or slice, not {type(index)}")

    def add(self, addr, name, type_key="b", display_key="hex"):
        # print("PROCESSING", addr, name, type_key, display_key, self.next_index)
        size_code = size_codes[type_key[-1]]
        try:
            item_count = int(type_key[:-1])
        except ValueError:
            item_count = 1
        display_code = display_codes[display_key]
        num_bytes = item_count * (size_code + 1)

        index = self.next_index
        r = self.labels['labels'][0][index]
        name = name.encode('utf-8')
        text_length = len(name)
        r['text_length'] = text_length
        r['num_bytes'] = num_bytes
        r['item_count'] = item_count
        r['type_code'] = size_code | display_code
        r['label'][0:text_length] = [c for c in name]

        self.labels['index'][0][addr] = index

        # long labels take over the next record entries
        if text_length > 12:
            text_length -= 12
            record_count = 1 + (text_length + 15) // 16
        else:
            record_count = 1
        self.next_index += record_count



import os
import inspect
import pkg_resources

from . import errors

import logging
log = logging.getLogger(__name__)


class Stringifier:
    """Stringifier (pretty printer) for data.

    Converts data to and from a human-readable representation.

    Stringifiers are stateless, but are implemented as normal classes for
    convenience and ease of subclassing.
    """
    text_type = None
    ui_name = "_base_"

    def __init__(self, byte_data=None):
        if byte_data is not None:
            self.text = self.calc_text(byte_data)

    def __str__(self):
        return self.text_type

    #### stringification

    def calc_text(self, byte_data):
        """Convert byte values into a string using this algorithm.

        `byte_data` must be a byte array or something that can pose for a byte
        array.

        If the data can't be losslessly represented by this stringifier, raise
        an `InvalidAlgorithm` exception. The caller must recognize this and
        take appropriate action.

        Subclasses should raise the `UnsupportedAlgorithm` exception if
        stringification is not implemented.
        """
        raise errors.UnsupportedAlgorithm(f"Stringifier '{self.text_type}' not implemented")

    #### parse

    def calc_byte_data(self, text_data):
        """Parse text using this stringifier's format.

        An `InvalidAlgorithm` exception will be raised if the text can't be
        parsed using this method.

        Subclasses should raise the `UnsupportedAlgorithm` exception if
        parsing is not implemented.
        """
        raise errors.UnsupportedAlgorithm(f"Parser '{self.text_type}' not implemented")


class ReprStringifier(Stringifier):
    text_type = "repr"
    ui_name = "Escaped String"

    def calc_text(self, byte_data):
        text = repr(byte_data)  # will generate byte string
        q = text[1]
        text = text[2:-1]  # remove 'b' and leading/trailing quotes
        if q == "'":
            # double quotes are literal, single quotes are escaped
            text = text.replace('"', "\\x22").replace("\\'", "\\x27")
        else:
            # single quotes are literal, double are escaped
            text = text.replace("'", "\\x27").replace('\\"', "\\x22")
        return text


_stringifiers = None

def _find_stringifiers():
    stringifiers = []
    for entry_point in pkg_resources.iter_entry_points('atrip.stringifiers'):
        mod = entry_point.load()
        log.debug(f"find_stringifier: Found module {entry_point.name}={mod.__name__}")
        for name, obj in inspect.getmembers(mod):
            if inspect.isclass(obj) and Stringifier in obj.__mro__[1:]:
                log.debug(f"find_stringifiers:   found stringifier class {name}")
                stringifiers.append(obj)
    stringifiers.append(ReprStringifier)
    return stringifiers

def find_stringifiers():
    global _stringifiers

    if _stringifiers is None:
        _stringifiers = _find_stringifiers()
    return _stringifiers

def find_stringifier_by_name(name):
    items = find_stringifiers()
    for c in items:
        if c.text_type == name:
            return c()
    raise KeyError(f"Unknown stringifier {name}")



import numpy as np

from .. import errors
from ..media_type import CartImage

import logging
log = logging.getLogger(__name__)


# From atari800 source
known_cart_types = [
# (note: all size units in KB)
# atari800 index number
# name
# total size
# static size
# static offset
# static address
# banked size
# banked offset (for bank zero)
# banked address
    (0,  "", 0,),
    (57, "Standard 2 KB", 2, 2, 0, 0xb800),
    (58, "Standard 4 KB", 4, 4, 0, 0xb000),
    (59, "Right slot 4 KB", 4, 4, 0, 0, 0x9000),
    (1,  "Standard 8 KB", 8, 8, 0, 0xa000),
    (21, "Right slot 8 KB", 8,),
    (2,  "Standard 16 KB", 16, 16, 0, 0x8000),
    (44, "OSS 8 KB", 8,),
    (15, "OSS one chip 16 KB", 16,),
    (3,  "OSS two chip (034M) 16 KB", 16, 4, 12, 0xb000, 4, 0, 0xa000),
    (45, "OSS two chip (043M) 16 KB", 16, 4, 12, 0xb000, 4, 0, 0xa000),
    (12, "XEGS 32 KB", 32,  8, 24, 0xa000, 8, 0, 0x8000),
    (13, "XEGS (banks 0-7) 64 KB", 64, 8, 56, 0xa000, 8, 0, 0x8000),
    (67, "XEGS (banks 8-15) 64 KB", 64, 8, 56, 0xa000, 8, 0, 0x8000),
    (14, "XEGS 128 KB", 128, 8, 120, 0xa000, 8, 0, 0x8000),
    (23, "XEGS 256 KB", 256, 8, 248, 0xa000, 8, 0, 0x8000),
    (24, "XEGS 512 KB", 512, 8, 504, 0xa000, 8, 0, 0x8000),
    (25, "XEGS 1 MB", 1024, 8, 1016, 0xa000, 8, 0, 0x8000 ),
    (33, "Switchable XEGS 32 KB", 32,  8, 24, 0xa000, 8, 0, 0x8000),
    (34, "Switchable XEGS 64 KB", 64,  8, 56, 0xa000, 8, 0, 0x8000),
    (35, "Switchable XEGS 128 KB", 128, 8, 120, 0xa000, 8, 0, 0x8000),
    (36, "Switchable XEGS 256 KB", 256, 8, 248, 0xa000, 8, 0, 0x8000),
    (37, "Switchable XEGS 512 KB", 512, 8, 504, 0xa000, 8, 0, 0x8000),
    (38, "Switchable XEGS 1 MB", 1024, 8, 1016, 0xa000, 8, 0, 0x8000 ),
    (22, "Williams 32 KB", 32,),
    (8,  "Williams 64 KB", 64,),
    (9,  "Express 64 KB", 64,),
    (10, "Diamond 64 KB", 64,),
    (11, "SpartaDOS X 64 KB", 64,),
    (43, "SpartaDOS X 128 KB", 128,),
    (17, "Atrax 128 KB", 128,),
    (18, "Bounty Bob 40 KB", 40,),
    (26, "MegaCart 16 KB", 16,),
    (27, "MegaCart 32 KB", 32,),
    (28, "MegaCart 64 KB", 64,),
    (29, "MegaCart 128 KB", 128,),
    (30, "MegaCart 256 KB", 256,),
    (31, "MegaCart 512 KB", 512,),
    (32, "MegaCart 1 MB", 1024,),
    (39, "Phoenix 8 KB", 8,),
    (46, "Blizzard 4 KB", 4,),
    (40, "Blizzard 16 KB", 16, 16, 0, 0x8000),
    (60, "Blizzard 32 KB", 32,),
    (41, "Atarimax 128 KB Flash", 128,),
    (42, "Atarimax 1 MB Flash", 1024,),
    (47, "AST 32 KB", 32,),
    (48, "Atrax SDX 64 KB", 64,),
    (49, "Atrax SDX 128 KB", 128,),
    (50, "Turbosoft 64 KB", 64,),
    (51, "Turbosoft 128 KB", 128,),
    (52, "Ultracart 32 KB", 32,),
    (53, "Low bank 8 KB", 8, 8, 0, 0x8000),
    (5,  "DB 32 KB", 32,),
    (54, "SIC! 128 KB", 128,),
    (55, "SIC! 256 KB", 256,),
    (56, "SIC! 512 KB", 512,),
    (61, "MegaMax 2 MB", 2048,),
    (62, "The!Cart 128 MB", 128*1024,),
    (63, "Flash MegaCart 4 MB", 4096,),
    (64, "MegaCart 2 MB", 2048,),
    (65, "The!Cart 32 MB", 32*1024,),
    (66, "The!Cart 64 MB", 64*1024,),
    (20, "Standard 4 KB 5200", 4, 4, 0, 0x8000),
    (19, "Standard 8 KB 5200", 8, 8, 0, 0x8000),
    (4,  "Standard 32 KB 5200", 32, 32, 0, 0x4000),
    (16, "One chip 16 KB 5200", 16,),
    (6,  "Two chip 16 KB 5200", 16,),
    (7,  "Bounty Bob 40 KB 5200", 40,),
]

known_cart_type_map = {c[0]:i for i, c in enumerate(known_cart_types)}


def get_known_carts():
    grouped = defaultdict(list)
    for c in known_cart_types[1:]:
        size = c[2]
        grouped[size].append(c)
    return grouped


def get_known_carts_of_size(kb):
    matches = []
    for c in known_cart_types[1:]:
        size = c[2]
        if size == kb:
            matches.append(c)
    return matches


def get_cart(cart_type):
    try:
        return known_cart_types[known_cart_type_map[cart_type]]
    except KeyError:
        raise errors.InvalidHeader("Unsupported cart type %d" % cart_type)


class A8CartHeader:
    # Atari Cart format described by https://sourceforge.net/p/atari800/source/ci/master/tree/DOC/cart.txt  NOTE: Big endian!
    format = np.dtype([
        ('magic', '|S4'),
        ('format', '>u4'),
        ('checksum', '>u4'),
        ('unused','>u4')
        ])

    def __init__(self, data):
        if len(data) == 16:
            header = data.view(dtype=self.format)[0]
            if header[0] != b'CART':
                raise errors.InvalidHeader("No Atari 8-bit cart header")
            self.cart_type = int(header[1])
            self.crc = int(header[2])
            self.set_type(self.cart_type)
        else:
            raise errors.InvalidHeader(f"Short cart header: should be 16 bytes; found {len(data)}")

    def __str__(self):
        return "%s Cartridge (atari800 type=%d size=%d, %d banks, crc=%d)" % (self.cart_name, self.cart_type, self.cart_size, self.bank_size, self.crc)

    def set_type(self, cart_type):
        self.cart_type = cart_type
        c = get_cart(cart_type)
        self.cart_name = c[1]
        self.cart_size = c[2]
        self.main_size = self.cart_size
        if len(c) >= 6:
            self.main_size, self.main_offset, self.main_origin = c[3:6]
        if len(c) >= 9:
            self.banks = []
            self.bank_size, offset, self.bank_origin = c[6:9]
            s = self.cart_size - self.main_size
            while s > 0:
                self.banks.append(offset)
                offset += self.bank_size
                s -= self.bank_size

    def check_media(self, media):
        media_size = len(media) - 16
        if self.cart_size != media_size:
            raise errors.InvalidHeader("Invalid cart size: {media_size}, expected {self.cart_size} for {self.cart_name}")


class Atari8bitCart(CartImage):
    ui_name = "Atari 8bit Cart"
    platform = "atari800"

    def calc_header(self, container):
        header_data = container[0:16]
        try:
            if len(header_data) == 16:
                header = A8CartHeader(header_data)
                header_length = 16
                header.check_media(container)
            else:
                header = None
        except errors.InvalidHeader:
            header = None
        return header

    def check_magic(self):
        s = self.signature
        if s is not None:
            if ".type" in s.mime:
                _, num = s.mime.split(".type")
                self.cart_type = int(num)
            if "atari800" in s.mime:
                self.platform = "atari800"
            elif "atari5200" in s.mime:
                self.platform = "atari5200"
        else:
            c = get_known_carts_of_size(self.kb)
            if not c:
                raise errors.InvalidMediaSize(f"No known {self.ui_name} of size {self.kb}k")
            guess = c[0]
            self.cart_type = guess[0]
            if "5200" in guess[1]:
                self.platform = "atari5200"
            else:
                self.platform = "atari800"



import numpy as np

import wx

from atrip import style_bits

from sawx.utils.permute import bit_reverse_table
from sawx.utils.nputil import intscale, intwscale, intwscale_font

from atrip.machines import atari8bit
from . import colors
try:
    from . import antic_speedups as speedups
except ImportError:
    speedups = None

import logging
log = logging.getLogger(__name__)


class BaseRenderer(object):
    name = "base"
    scale_width = 1
    scale_height = 1
    pixels_per_byte = 8
    bitplanes = 1
    ignore_mask = style_bits.not_user_bit_mask & (0xff ^ style_bits.diff_bit_mask)

    def __str__(self):
        return self.name

    def __eq__(self, other):
        try:
            return other is not None and self.name == other.name and self.pixels_per_byte == other.pixels_per_byte and self.bitplanes == other.bitplanes
        except AttributeError:
            pass
        return False

    # to be usable in dicts, py3 needs __hash__ defined if __eq__ is defined
    def __hash__(self):
        return id(self)

    def validate_bytes_per_row(self, bytes_per_row):
        return bytes_per_row

    def get_colors(self, segment_viewer, registers):
        color_registers = [segment_viewer.color_registers[r] for r in registers]
        log.debug(f"get_colors: {color_registers} from {segment_viewer}")
        h_colors = colors.get_blended_color_registers(color_registers, segment_viewer.preferences.highlight_background_color)
        m_colors = colors.get_blended_color_registers(color_registers, segment_viewer.preferences.match_background_color)
        c_colors = colors.get_blended_color_registers(color_registers, segment_viewer.preferences.comment_background_color)
        d_colors = colors.get_dimmed_color_registers(color_registers, segment_viewer.preferences.background_color, segment_viewer.preferences.data_background_color)
        return color_registers, h_colors, m_colors, c_colors, d_colors

    def reshape(self, bitimage, bytes_per_row, nr):
        # source array 'bitimage' in the shape of (size, w, 3)
        h, w, colors = bitimage.shape
        # create a new image with pixels in the correct aspect ratio
        output = bitimage.reshape((nr, self.pixels_per_byte * bytes_per_row, 3))
        output = intscale(output, self.scale_height, self.scale_width)
        log.debug("bitimage: %d,%d,%d; ppb=%d bpr=%d, output=%s" % (h, w, colors, self.pixels_per_byte * self.scale_width, bytes_per_row, str(output.shape)))
        return output

    def calc_style_per_pixel_1bpp(self, style):
        stack = np.empty((len(style), 8), dtype=style.dtype)
        stack[:,0] = style
        stack[:,1] = style
        stack[:,2] = style
        stack[:,3] = style
        stack[:,4] = style
        stack[:,5] = style
        stack[:,6] = style
        stack[:,7] = style
        return stack

    def calc_style_per_pixel_2bpp(self, style):
        stack = np.empty((len(style), 4), dtype=style.dtype)
        stack[:,0] = style
        stack[:,1] = style
        stack[:,2] = style
        stack[:,3] = style
        return stack

    def calc_style_per_pixel_4bpp(self, style):
        stack = np.empty((len(style), 2), dtype=style.dtype)
        stack[:,0] = style
        stack[:,1] = style
        return stack

    def calc_style_per_pixel(self, style):
        if self.pixels_per_byte == 1:
            return style
        elif self.pixels_per_byte == 2:
            return self.calc_style_per_pixel_4bpp(style)
        elif self.pixels_per_byte == 4:
            return self.calc_style_per_pixel_2bpp(style)
        elif self.pixels_per_byte == 8:
            return self.calc_style_per_pixel_1bpp(style)

    def get_2bpp(self, segment_viewer, bytes_per_row, nr, count, byte_values, style, colors, style_per_pixel=None):
        bits = np.unpackbits(byte_values)
        bits = bits.reshape((-1, 8))
        pixels = np.empty((nr * bytes_per_row, 4), dtype=np.uint8)
        pixels[:,0] = bits[:,0] * 2 + bits[:,1]
        pixels[:,1] = bits[:,2] * 2 + bits[:,3]
        pixels[:,2] = bits[:,4] * 2 + bits[:,5]
        pixels[:,3] = bits[:,6] * 2 + bits[:,7]

        if style_per_pixel is None:
            style_per_pixel = self.calc_style_per_pixel_2bpp(style)
        normal = (style_per_pixel & self.ignore_mask) == 0
        highlight = (style_per_pixel & style_bits.selected_bit_mask) == style_bits.selected_bit_mask
        data = (style_per_pixel & style_bits.data_bit_mask) == style_bits.data_bit_mask
        comment = (style_per_pixel & style_bits.comment_bit_mask) == style_bits.comment_bit_mask
        match = (style_per_pixel & style_bits.match_bit_mask) == style_bits.match_bit_mask

        color_registers, h_colors, m_colors, c_colors, d_colors = colors
        bitimage = np.empty((nr * bytes_per_row, 4, 3), dtype=np.uint8)
        for i in range(4):
            color_is_set = (pixels == i)
            bitimage[color_is_set & normal] = color_registers[i]
            bitimage[color_is_set & data] = d_colors[i]
            bitimage[color_is_set & comment] = c_colors[i]
            bitimage[color_is_set & match] = m_colors[i]
            bitimage[color_is_set & highlight] = h_colors[i]
        bitimage[count:,:,:] = segment_viewer.preferences.empty_background_color.Get(False)
        return bitimage

    def get_4bpp(self, segment_viewer, bytes_per_row, nr, count, byte_values, style, colors, style_per_pixel=None):
        bits = np.unpackbits(byte_values)
        bits = bits.reshape((-1, 8))
        pixels = np.empty((nr * bytes_per_row, 2), dtype=np.uint8)
        pixels[:,0] = bits[:,0] * 8 + bits[:,1] * 4 + bits[:,2] * 2 + bits[:,3]
        pixels[:,1] = bits[:,4] * 8 + bits[:,5] * 4 + bits[:,6] * 2 + bits[:,7]

        if style_per_pixel is None:
            style_per_pixel = self.calc_style_per_pixel_4bpp(style)
        normal = (style_per_pixel & self.ignore_mask) == 0
        highlight = (style_per_pixel & style_bits.selected_bit_mask) == style_bits.selected_bit_mask
        data = (style_per_pixel & style_bits.data_bit_mask) == style_bits.data_bit_mask
        comment = (style_per_pixel & style_bits.comment_bit_mask) == style_bits.comment_bit_mask
        match = (style_per_pixel & style_bits.match_bit_mask) == style_bits.match_bit_mask

        color_registers, h_colors, m_colors, c_colors, d_colors = colors
        bitimage = np.empty((nr * bytes_per_row, 2, 3), dtype=np.uint8)
        for i in range(16):
            color_is_set = (pixels == i)
            bitimage[color_is_set & normal] = color_registers[i]
            bitimage[color_is_set & data] = d_colors[i]
            bitimage[color_is_set & comment] = c_colors[i]
            bitimage[color_is_set & match] = m_colors[i]
            bitimage[color_is_set & highlight] = h_colors[i]
        bitimage[count:,:,:] = segment_viewer.preferences.empty_background_color.Get(False)
        return bitimage

    def get_bitplane_pixels(self, bits, pixels, bytes_per_row, pixels_per_row):
        """Fill the pixels array with color register data
        
        pixels is passed in as an 8 pixel wide array regardless of the actual
        pixels per row. It will be resized correctly by the calling method.
        """
        raise NotImplemented

    def get_bitplane_style(self, style):
        raise NotImplemented

    def get_bitplanes(self, segment_viewer, bytes_per_row, nr, count, byte_values, style, colors):
        bitplanes = self.bitplanes
        _, rem = divmod(np.alen(byte_values), bitplanes)
        if rem > 0:
            byte_values = np.append(byte_values, np.zeros(rem, dtype=np.uint8))
            style = np.append(style, np.zeros(rem, dtype=np.uint8))
        pixels_per_row = 8 * bytes_per_row // bitplanes
        bits = np.unpackbits(byte_values).reshape((-1, 8))
        pixels = np.empty((nr * bytes_per_row // bitplanes, pixels_per_row), dtype=np.uint8)
        self.get_bitplane_pixels(bits, pixels, bytes_per_row, pixels_per_row)
        pixels = pixels.reshape((nr, pixels_per_row))
        s = self.get_bitplane_style(style)
        style_per_pixel = s.repeat(8).reshape((-1, pixels_per_row))
        normal = (style_per_pixel & self.ignore_mask) == 0
        highlight = (style_per_pixel & style_bits.selected_bit_mask) == style_bits.selected_bit_mask
        data = (style_per_pixel & style_bits.data_bit_mask) == style_bits.data_bit_mask
        comment = (style_per_pixel & style_bits.comment_bit_mask) == style_bits.comment_bit_mask
        match = (style_per_pixel & style_bits.match_bit_mask) == style_bits.match_bit_mask

        color_registers, h_colors, m_colors, c_colors, d_colors = colors
        bitimage = np.empty((nr, pixels_per_row, 3), dtype=np.uint8)
        for i in range(2**bitplanes):
            color_is_set = (pixels == i)
            bitimage[color_is_set & normal] = color_registers[i]
            bitimage[color_is_set & data] = d_colors[i]
            bitimage[color_is_set & comment] = c_colors[i]
            bitimage[color_is_set & match] = m_colors[i]
            bitimage[color_is_set & highlight] = h_colors[i]
        bitimage[count:,:,:] = segment_viewer.preferences.empty_background_color.Get(False)
        return bitimage


class OneBitPerPixelB(BaseRenderer):
    name = "B/W, 1bpp, on=black"

    def get_bw_colors(self, segment_viewer):
        return ((255, 255, 255), (0, 0, 0))

    def get_image(self, segment_viewer, bytes_per_row, nr, count, byte_values, style, style_per_pixel=None):
        bits = np.unpackbits(byte_values)
        pixels = bits.reshape((-1, 8))

        background = (pixels == 0)
        color1 = (pixels == 1)

        bw_colors = self.get_bw_colors(segment_viewer)
        h_colors = colors.get_blended_color_registers(bw_colors, segment_viewer.preferences.highlight_background_color)
        m_colors = colors.get_blended_color_registers(bw_colors, segment_viewer.preferences.match_background_color)
        c_colors = colors.get_blended_color_registers(bw_colors, segment_viewer.preferences.comment_background_color)
        d_colors = colors.get_dimmed_color_registers(bw_colors, segment_viewer.preferences.background_color, segment_viewer.preferences.data_background_color)

        if style_per_pixel is None:
            style_per_pixel = self.calc_style_per_pixel_1bpp(style)
        normal = (style_per_pixel & self.ignore_mask) == 0
        highlight = (style_per_pixel & style_bits.selected_bit_mask) == style_bits.selected_bit_mask
        data = (style_per_pixel & style_bits.data_bit_mask) == style_bits.data_bit_mask
        comment = (style_per_pixel & style_bits.comment_bit_mask) == style_bits.comment_bit_mask
        match = (style_per_pixel & style_bits.match_bit_mask) == style_bits.match_bit_mask

        bitimage = np.empty((nr * bytes_per_row, 8, 3), dtype=np.uint8)
        bitimage[background & normal] = bw_colors[0]
        bitimage[background & data] = d_colors[0]
        bitimage[background & comment] = c_colors[0]
        bitimage[background & match] = m_colors[0]
        bitimage[background & highlight] = h_colors[0]
        bitimage[color1 & normal] = bw_colors[1]
        bitimage[color1 & data] = d_colors[1]
        bitimage[color1 & comment] = c_colors[1]
        bitimage[color1 & match] = m_colors[1]
        bitimage[color1 & highlight] = h_colors[1]
        bitimage[count:,:,:] = segment_viewer.preferences.empty_background_color.Get(False)

        return self.reshape(bitimage, bytes_per_row, nr)


class OneBitPerPixelW(OneBitPerPixelB):
    name = "B/W, 1bpp, on=white"

    def get_bw_colors(self, segment_viewer):
        return ((0, 0, 0), (255, 255, 255))


class OneBitPerPixelPM1(OneBitPerPixelB):
    name = "Player/Missile, normal width"
    scale_width = 2


class OneBitPerPixelPM2(OneBitPerPixelB):
    name = "Player/Missile, double width"
    scale_width = 4


class OneBitPerPixelPM4(OneBitPerPixelB):
    name = "Player/Missile, quad width"
    scale_width = 8


class ModeB(OneBitPerPixelB):
    name = "Antic B (Gr 6, 1bpp)"
    scale_width = 2
    scale_height = 2

    def get_bw_colors(self, segment_viewer):
        color_registers = [segment_viewer.color_registers[r] for r in [8, 4]]
        return color_registers


class ModeC(ModeB):
    name = "Antic C (Gr 6+, 1bpp)"
    scale_width = 2
    scale_height = 2


class OneBitPerPixelApple2Linear(BaseRenderer):
    name = "B/W, Apple 2, Linear"

    def get_bw_colors(self, segment_viewer):
        return ((0, 0, 0), (255, 255, 255))

    def calc_style_per_pixel(self, style):
        return np.vstack((style, style, style, style, style, style, style)).T

    def get_image(self, segment_viewer, bytes_per_row, nr, count, byte_values, style, style_per_pixel=None):
        bits = np.unpackbits(bit_reverse_table[byte_values])
        pixels = bits.reshape((-1, 8))

        background = (pixels[:,0:7] == 0)
        color1 = (pixels[:,0:7] == 1)

        bw_colors = self.get_bw_colors(segment_viewer)
        h_colors = colors.get_blended_color_registers(bw_colors, segment_viewer.preferences.highlight_background_color)
        m_colors = colors.get_blended_color_registers(bw_colors, segment_viewer.preferences.match_background_color)
        c_colors = colors.get_blended_color_registers(bw_colors, segment_viewer.preferences.comment_background_color)
        d_colors = colors.get_dimmed_color_registers(bw_colors, segment_viewer.preferences.background_color, segment_viewer.preferences.data_background_color)

        if style_per_pixel is None:
            style_per_pixel = self.calc_style_per_pixel(style)
        normal = (style_per_pixel & self.ignore_mask) == 0
        highlight = (style_per_pixel & style_bits.selected_bit_mask) == style_bits.selected_bit_mask
        data = (style_per_pixel & style_bits.data_bit_mask) == style_bits.data_bit_mask
        comment = (style_per_pixel & style_bits.comment_bit_mask) == style_bits.comment_bit_mask
        match = (style_per_pixel & style_bits.match_bit_mask) == style_bits.match_bit_mask

        bitimage = np.empty((nr * bytes_per_row, 7, 3), dtype=np.uint8)
        bitimage[background & normal] = bw_colors[0]
        bitimage[background & data] = d_colors[0]
        bitimage[background & comment] = c_colors[0]
        bitimage[background & match] = m_colors[0]
        bitimage[background & highlight] = h_colors[0]
        bitimage[color1 & normal] = bw_colors[1]
        bitimage[color1 & data] = d_colors[1]
        bitimage[color1 & comment] = c_colors[1]
        bitimage[color1 & match] = m_colors[1]
        bitimage[color1 & highlight] = h_colors[1]
        bitimage[count:,:,:] = segment_viewer.preferences.empty_background_color.Get(False)

        return bitimage.reshape((nr, bytes_per_row * 7, 3))


def generate_apple2_row_offsets():
    offsets = np.zeros(192, dtype=np.int32)
    for y in range(192):
        # From Apple Graphics and Arcade Game Design
        a = y // 64
        d = y - (64 * a)
        b = d // 8
        c = d - 8 * b
        offsets[i] = (1024 * c) + (128 * b) + (40 * a)
    return offsets

def generate_apple2_index():
    offsets = generate_apple2_row_offsets()
    bytepos = np.empty((192, 280), dtype=np.int32)
    bytepos[:,0] = offsets * 7

class OneBitPerPixelApple2FullScreen(OneBitPerPixelApple2Linear):
    name = "B/W, Apple 2, Screen Order"

    def get_image(self, segment_viewer, bytes_per_row, nr, count, byte_values, style, style_per_pixel=None):
        screen = np.zeros((8192,), dtype=np.uint8)
        byte_values = byte_values[0:8192]
        num_valid = len(byte_values)  # might be smaller than 8192
        screen[:num_valid] = byte_values
        bits = np.unpackbits(bit_reverse_table[screen])
        pixels = bits.reshape((-1, 8))

        background = (pixels[:,0:7] == 0)
        color1 = (pixels[:,0:7] == 1)

        bw_colors = self.get_bw_colors(segment_viewer)
        h_colors = colors.get_blended_color_registers(bw_colors, segment_viewer.preferences.highlight_background_color)
        m_colors = colors.get_blended_color_registers(bw_colors, segment_viewer.preferences.match_background_color)
        c_colors = colors.get_blended_color_registers(bw_colors, segment_viewer.preferences.comment_background_color)
        d_colors = colors.get_dimmed_color_registers(bw_colors, segment_viewer.preferences.background_color, segment_viewer.preferences.data_background_color)

        if style_per_pixel is None:
            style_per_pixel = self.calc_style_per_pixel(style)
        normal = (style_per_pixel & self.ignore_mask) == 0
        highlight = (style_per_pixel & style_bits.selected_bit_mask) == style_bits.selected_bit_mask
        data = (style_per_pixel & style_bits.data_bit_mask) == style_bits.data_bit_mask
        comment = (style_per_pixel & style_bits.comment_bit_mask) == style_bits.comment_bit_mask
        match = (style_per_pixel & style_bits.match_bit_mask) == style_bits.match_bit_mask

        bitimage = np.empty((192 * 40, 7, 3), dtype=np.uint8)
        bitimage[background & normal] = bw_colors[0]
        bitimage[background & data] = d_colors[0]
        bitimage[background & comment] = c_colors[0]
        bitimage[background & match] = m_colors[0]
        bitimage[background & highlight] = h_colors[0]
        bitimage[color1 & normal] = bw_colors[1]
        bitimage[color1 & data] = d_colors[1]
        bitimage[color1 & comment] = c_colors[1]
        bitimage[color1 & match] = m_colors[1]
        bitimage[color1 & highlight] = h_colors[1]
        bitimage[count:,:,:] = segment_viewer.preferences.empty_background_color.Get(False)

        return bitimage.reshape((nr, bytes_per_row * 7, 3))


class OneBitPerPixelApple2Artifacting(OneBitPerPixelApple2Linear):
    name = "Apple 2 (artifacting colors)"

    def get_bw_colors(self, segment_viewer):
        return ((255, 255, 255), (0, 0, 0))

    # 0 0000000 0 0000000  # black 0, 0, 0
    # 0 0101010 0 1010101  # green 32, 192, 0
    # 0 1010101 0 0101010  # purple 159, 0, 253
    # 0 1111111 0 1111111  # white 255, 255, 255
    # 1 0000000 1 0000000  # black 0, 0, 0
    # 1 0101010 1 1010101  # orange 240, 80, 0
    # 1 1010101 1 0101010  # blue 0, 128, 255
    # 1 1111111 1 1111111  # white 255, 255, 255
    #
    # From https://groups.google.com/forum/#!msg/comp.sys.apple2.programmer/vxtFo6QEYGg/9kh1RfovGQAJ
    # bit #
    # 0 1 2 3 4 5 6 7   0 1 2 3 4 5 6 7

    #          color bit          color bit
    #               |                 |
    # 0 0 0 0 0 0 0 0   0 0 0 0 0 0 0 0
    # |_| |_| |_| |_____| |_| |_| |_|

    #  0   1   2     3     4   5   6

    # every 2 byte_values starting on an even number ($2000.2001) gives seven pixels.
    # As you can see, pixel #3 uses a bit from byte $2000 and byte 2001 to form
    # a pixel. if any two bits are consecutive (except the color bit) you get
    # two white pixels. with color bit clear, each group of bits will show
    # violet or green. with color bit set, each group of bits will show blue or
    # orange.

    # color bit clear
    # bits-color
    # 00 - black
    # 10 - violet
    # 01 - green
    # 11 - white

    # color bit set
    # 00 - black
    # 10 - blue
    # 01 - orange
    # 11 - white

    def get_image(self, segment_viewer, bytes_per_row, nr, count, byte_values, style, style_per_pixel=None):
        bits = np.unpackbits(bit_reverse_table[byte_values])
        pixels = bits.reshape((-1, 8))

        background = (pixels[:,0:7] == 0)
        color1 = (pixels[:,0:7] == 1)

        bw_colors = self.get_bw_colors(segment_viewer)
        h_colors = colors.get_blended_color_registers(bw_colors, segment_viewer.preferences.highlight_background_color)
        m_colors = colors.get_blended_color_registers(bw_colors, segment_viewer.preferences.match_background_color)
        c_colors = colors.get_blended_color_registers(bw_colors, segment_viewer.preferences.comment_background_color)
        d_colors = colors.get_dimmed_color_registers(bw_colors, segment_viewer.preferences.background_color, segment_viewer.preferences.data_background_color)

        if style_per_pixel is None:
            style_per_pixel = self.calc_style_per_pixel(style)
        normal = (style_per_pixel & self.ignore_mask) == 0
        highlight = (style_per_pixel & style_bits.selected_bit_mask) == style_bits.selected_bit_mask
        data = (style_per_pixel & style_bits.data_bit_mask) == style_bits.data_bit_mask
        comment = (style_per_pixel & style_bits.comment_bit_mask) == style_bits.comment_bit_mask
        match = (style_per_pixel & style_bits.match_bit_mask) == style_bits.match_bit_mask

        bitimage = np.empty((nr * bytes_per_row, 7, 3), dtype=np.uint8)
        bitimage[background & normal] = bw_colors[0]
        bitimage[background & data] = d_colors[0]
        bitimage[background & comment] = c_colors[0]
        bitimage[background & match] = m_colors[0]
        bitimage[background & highlight] = h_colors[0]
        bitimage[color1 & normal] = bw_colors[1]
        bitimage[color1 & data] = d_colors[1]
        bitimage[color1 & comment] = c_colors[1]
        bitimage[color1 & match] = m_colors[1]
        bitimage[color1 & highlight] = h_colors[1]
        bitimage[count:,:,:] = segment_viewer.preferences.empty_background_color.Get(False)

        return bitimage.reshape((nr, bytes_per_row * 7, 3))


class TwoBitsPerPixel(BaseRenderer):
    name = "2bpp"
    pixels_per_byte = 4

    def get_image(self, segment_viewer, bytes_per_row, nr, count, byte_values, style, style_per_pixel=None):
        colors = self.get_colors(segment_viewer, [0, 1, 2, 3])
        bitimage = self.get_2bpp(segment_viewer, bytes_per_row, nr, count, byte_values, style, colors, style_per_pixel)
        return self.reshape(bitimage, bytes_per_row, nr)


class ModeD(TwoBitsPerPixel):
    name = "Antic D (Gr 7, 2bpp)"
    scale_width = 2
    scale_height = 2
    pixels_per_byte = 4

    def get_image(self, segment_viewer, bytes_per_row, nr, count, byte_values, style, style_per_pixel=None):
        colors = self.get_colors(segment_viewer, [8, 4, 5, 6])
        bitimage = self.get_2bpp(segment_viewer, bytes_per_row, nr, count, byte_values, style, colors, style_per_pixel)
        return self.reshape(bitimage, bytes_per_row, nr)


class ModeE(ModeD):
    name = "Antic E (Gr 7+, 2bpp)"
    scale_width = 2
    scale_height = 1
    pixels_per_byte = 4


class FourBitsPerPixel(BaseRenderer):
    name = "4bpp"
    pixels_per_byte = 2

    def get_image(self, segment_viewer, bytes_per_row, nr, count, byte_values, style, style_per_pixel=None):
        colors = self.get_colors(segment_viewer, list(range(16)))
        bitimage = self.get_4bpp(segment_viewer, bytes_per_row, nr, count, byte_values, style, colors, style_per_pixel)
        return self.reshape(bitimage, bytes_per_row, nr)


class TwoBitPlanesLE(BaseRenderer):
    name = "2 Bit Planes (little endian)"
    pixels_per_byte = 8
    bitplanes = 2

    def get_bitplane_pixels(self, bits, pixels, bytes_per_row, pixels_per_row):
        for i in range(pixels_per_row):
            pixels[:,i] = bits[0::2,i] + bits[1::2,i] * 2

    def get_bitplane_style(self, style):
        return style[0::2] | style[1::2]

    def validate_bytes_per_row(self, bytes_per_row):
        scale, rem = divmod(bytes_per_row, self.bitplanes)
        if rem > 0:
            bytes_per_row = (scale + 1) * self.bitplanes
        return bytes_per_row

    def get_image(self, segment_viewer, bytes_per_row, nr, count, byte_values, style, style_per_pixel=None):
        colors = self.get_colors(segment_viewer, list(range(2**self.bitplanes)))
        bitimage = self.get_bitplanes(segment_viewer, bytes_per_row, nr, count, byte_values, style, colors, style_per_pixel)
        return bitimage


class TwoBitPlanesBE(TwoBitPlanesLE):
    name = "2 Bit Planes (big endian)"

    def get_bitplane_pixels(self, bits, pixels, bytes_per_row, pixels_per_row):
        for i in range(pixels_per_row):
            pixels[:,i] = bits[0::2,i] * 2 + bits[1::2,i]


class TwoBitPlanesLineLE(TwoBitPlanesLE):
    name = "2 Bit Planes (little endian, interleave by line)"

    def get_bitplane_pixels(self, bits, pixels, bytes_per_row, pixels_per_row):
        pixel_rows = bytes_per_row // 2
        for i in range(pixels_per_row):
            for j in range(pixel_rows):
                little = j
                big = j + pixel_rows
                pixels[j::pixel_rows,i] = bits[big::bytes_per_row,i] * 2 + bits[little::bytes_per_row,i]


class TwoBitPlanesLineBE(TwoBitPlanesLE):
    name = "2 Bit Planes (big endian, interleave by line)"

    def get_bitplane_pixels(self, bits, pixels, bytes_per_row, pixels_per_row):
        pixel_rows = bytes_per_row // 2
        for i in range(pixels_per_row):
            for j in range(pixel_rows):
                little = j + pixel_rows
                big = j
                pixels[j::pixel_rows,i] = bits[big::bytes_per_row,i] * 2 + bits[little::bytes_per_row,i]


class ThreeBitPlanesLE(TwoBitPlanesLE):
    name = "3 Bit Planes (little endian)"
    bitplanes = 3

    def get_bitplane_pixels(self, bits, pixels, bytes_per_row, pixels_per_row):
        for i in range(pixels_per_row):
            pixels[:,i] = bits[0::3,i] * 4 + bits[1::3,i] * 2 + bits[2::3,i]

    def get_bitplane_style(self, style):
        return style[0::3] | style[1::3] | style[2::3]


class ThreeBitPlanesBE(ThreeBitPlanesLE):
    name = "3 Bit Planes (big endian)"
    bitplanes = 3

    def get_bitplane_pixels(self, bits, pixels, bytes_per_row, pixels_per_row):
        for i in range(pixels_per_row):
            pixels[:,i] = bits[0::3,i] + bits[1::3,i] * 2 + bits[2::3,i] * 4


class ThreeBitPlanesLineLE(ThreeBitPlanesLE):
    name = "3 Bit Planes (little endian, interleave by line)"

    def get_bitplane_pixels(self, bits, pixels, bytes_per_row, pixels_per_row):
        pixel_rows = bytes_per_row // 3
        for i in range(pixels_per_row):
            for j in range(pixel_rows):
                little = j
                mid = j + pixel_rows
                big = j + (2 * pixel_rows)
                pixels[j::pixel_rows,i] = bits[big::bytes_per_row,i] * 4 + bits[mid::bytes_per_row,i] * 2 + bits[little::bytes_per_row,i]


class ThreeBitPlanesLineBE(ThreeBitPlanesBE):
    name = "3 Bit Planes (big endian, interleave by line)"

    def get_bitplane_pixels(self, bits, pixels, bytes_per_row, pixels_per_row):
        pixel_rows = bytes_per_row // 3
        for i in range(pixels_per_row):
            for j in range(pixel_rows):
                little = j + (2 * pixel_rows)
                mid = j + pixel_rows
                big = j
                pixels[j::pixel_rows,i] = bits[big::bytes_per_row,i] * 4 + bits[mid::bytes_per_row,i] * 2 + bits[little::bytes_per_row,i]


class FourBitPlanesLE(TwoBitPlanesLE):
    name = "4 Bit Planes (little endian)"
    bitplanes = 4

    def get_bitplane_pixels(self, bits, pixels, bytes_per_row, pixels_per_row):
        for i in range(pixels_per_row):
            pixels[:,i] = bits[0::4,i] * 8 + bits[1::4,i] * 4 + bits[2::4,i] * 2 + bits[3::4,i]

    def get_bitplane_style(self, style):
        return style[0::4] | style[1::4] | style[2::4] | style[3::4]


class FourBitPlanesBE(FourBitPlanesLE):
    name = "4 Bit Planes (big endian)"

    def get_bitplane_pixels(self, bits, pixels, bytes_per_row, pixels_per_row):
        for i in range(pixels_per_row):
            pixels[:,i] = bits[0::4,i] + bits[1::4,i] * 2 + bits[2::4,i] * 4 + bits[3::4,i] * 8


class FourBitPlanesLineLE(FourBitPlanesLE):
    name = "4 Bit Planes (little endian, interleave by line)"

    def get_bitplane_pixels(self, bits, pixels, bytes_per_row, pixels_per_row):
        pixel_rows = bytes_per_row // 4
        for i in range(pixels_per_row):
            for j in range(pixel_rows):
                little = j
                little_mid = j + pixel_rows
                big_mid = j + (2 * pixel_rows)
                big = j + (3 * pixel_rows)
                pixels[j::pixel_rows,i] = bits[big::bytes_per_row,i] * 8 + bits[big_mid::bytes_per_row,i] * 4 + bits[little_mid::bytes_per_row,i] * 2 + bits[little::bytes_per_row,i]


class FourBitPlanesLineBE(FourBitPlanesLE):
    name = "4 Bit Planes (big endian, interleave by line)"

    def get_bitplane_pixels(self, bits, pixels, bytes_per_row, pixels_per_row):
        pixel_rows = bytes_per_row // 4
        for i in range(pixels_per_row):
            for j in range(pixel_rows):
                little = j + (3 * pixel_rows)
                little_mid = j + (2 * pixel_rows)
                big_mid = j + pixel_rows
                big = j
                pixels[j::pixel_rows,i] = bits[big::bytes_per_row,i] * 8 + bits[big_mid::bytes_per_row,i] * 4 + bits[little_mid::bytes_per_row,i] * 2 + bits[little::bytes_per_row,i]


class GTIA9(FourBitsPerPixel):
    name = "GTIA 9 (4bpp, 16 luminances, 1 color)"
    scale_width = 4
    pixels_per_byte = 2

    def get_antic_color_registers(self, segment_viewer):
        first_color = segment_viewer.antic_color_registers[8] & 0xf0
        return list(range(first_color, first_color + 16))

    def get_colors(self, segment_viewer, registers):
        antic_color_registers = self.get_antic_color_registers(segment_viewer)
        color_registers = colors.get_color_registers(antic_color_registers, segment_viewer.color_standard)
        h_colors = colors.get_blended_color_registers(color_registers, segment_viewer.preferences.highlight_background_color)
        m_colors = colors.get_blended_color_registers(color_registers, segment_viewer.preferences.match_background_color)
        c_colors = colors.get_blended_color_registers(color_registers, segment_viewer.preferences.comment_background_color)
        d_colors = colors.get_dimmed_color_registers(color_registers, segment_viewer.preferences.background_color, segment_viewer.preferences.data_background_color)
        return color_registers, h_colors, m_colors, c_colors, d_colors


class GTIA10(GTIA9):
    name = "GTIA 10 (4bpp, 9 colors)"

    def get_antic_color_registers(self, segment_viewer):
        return list(segment_viewer.antic_color_registers[0:9]) + [0] * 7


class GTIA11(GTIA9):
    name = "GTIA 11 (4bpp, 1 luminace, 16 colors)"

    def get_antic_color_registers(self, segment_viewer):
        first_color = segment_viewer.antic_color_registers[8] & 0x0f
        return list(range(first_color, first_color + 256, 16))


class BaseBytePerPixelRenderer(BaseRenderer):
    """A generic renderer to display one pixel of the source using a byte
    for the image. This provides highlighting on a per-pixel level which isn't possible with modes that have multiple pixels in a byte. This is used as an intermediate renderer, so the image format can convert to byte per pixel, and this renderer can display it.

    Theoretically this will be able to display 256 colors, but at the
    moment only uses the first 16 colors. It is mapped to the ANTIC color
    register order, so the first 4 colors are player colors, then the 5
    playfield colors. A blank screen corresponds to the index value of 8, so
    the last playfield color.
    """
    name = "Intermediate Mode 1 Byte Per Pixel"
    pixels_per_byte = 1
    bitplanes = 1

    def pixels_from_2bpp(self, segment_viewer, bytes_per_row, nr, count, byte_values, style, colors):
        bits = np.unpackbits(byte_values)
        bits = bits.reshape((-1, 8))
        pixels = np.empty((nr * bytes_per_row, 4), dtype=np.uint8)
        pixels[:,0] = bits[:,0] * 2 + bits[:,1]
        pixels[:,1] = bits[:,2] * 2 + bits[:,3]
        pixels[:,2] = bits[:,4] * 2 + bits[:,5]
        pixels[:,3] = bits[:,6] * 2 + bits[:,7]
        style_per_pixel = np.vstack((style, style, style, style)).T
        return pixels, style_per_pixel

    def get_image(self, segment_viewer, bytes_per_row, nr, count, byte_values, style, style_per_pixel=None):
        normal = style == 0
        highlight = (style & style_bits.selected_bit_mask) == style_bits.selected_bit_mask
        comment = (style & style_bits.comment_bit_mask) == style_bits.comment_bit_mask
        data = (style & style_bits.data_bit_mask) == style_bits.data_bit_mask
        match = (style & style_bits.match_bit_mask) == style_bits.match_bit_mask

        color_registers, h_colors, m_colors, c_colors, d_colors = self.get_colors(segment_viewer, list(range(16)))
        bitimage = np.empty((nr * bytes_per_row, 3), dtype=np.uint8)
        for i in range(16):
            color_is_set = (byte_values == i)
            bitimage[color_is_set & normal] = color_registers[i]
            bitimage[color_is_set & comment] = c_colors[i]
            bitimage[color_is_set & match] = m_colors[i]
            bitimage[color_is_set & data] = d_colors[i]
            bitimage[color_is_set & highlight] = h_colors[i]
        bitimage[count:,:] = segment_viewer.preferences.empty_background_color.Get(False)
        return bitimage.reshape((nr, bytes_per_row, 3))


class BytePerPixelMemoryMap(BaseRenderer):
    name = "1Bpp Greyscale"

    def get_image(self, segment_viewer, byte_values, style, start_byte, end_byte, bytes_per_row, nr, start_col, visible_cols):
        if speedups is not None:
            array = speedups.get_numpy_memory_map_image(segment_viewer, byte_values, style, start_byte, end_byte, bytes_per_row, nr, start_col, visible_cols)
        else:
            array = get_numpy_memory_map_image(segment_viewer, byte_values, style, start_byte, end_byte, bytes_per_row, nr, start_col, visible_cols)
        return array

    def get_image(self, segment_viewer, bytes_per_row, nr, count, byte_values, style, style_per_pixel=None):
        array = get_numpy_memory_map_image(segment_viewer, bytes_per_row, nr, count, byte_values, style)
        return array


def get_numpy_memory_map_image(segment_viewer, bytes_per_row, nr, count, byte_values, style):
    log.debug("SLOW VERSION OF get_numpy_memory_map_image!!!")

    array = np.empty((nr, bytes_per_row, 3), dtype=np.uint8)
    array[:,:] = segment_viewer.preferences.empty_background_color.Get(False)
    selected_color = segment_viewer.preferences.highlight_background_color

    for j in range(nr):
        for i in range(bytes_per_row):
            c = byte_values[j * 256 + i] ^ 0xff
            s = style[j * 256 + i]
            if s & style_bits.selected_bit_mask:
                r = selected_color[0] * c >> 8
                g = selected_color[1] * c >> 8
                b = selected_color[2] * c >> 8
                array[j,i,:] = (r, g, b)
            else:
                array[j,i,:] = (c, c, c)
    return array


class MemoryAccessMap(BytePerPixelMemoryMap):
    name = "Memory Access"

    def get_image(self, segment_viewer, bytes_per_row, nr, count, byte_values, style):
        if speedups is not None:
            source = byte_values.reshape((nr, bytes_per_row))
            style = style.reshape((nr, bytes_per_row))
            array = speedups.get_numpy_memory_access_image(segment_viewer, bytes_per_row, nr, count, source, style, 0, bytes_per_row)
        else:
            array = get_numpy_memory_access_image(segment_viewer, bytes_per_row, nr, count, byte_values, style)
        return array


def get_numpy_memory_access_image(segment_viewer, bytes_per_row, nr, count, byte_values, style):

    source = byte_values.reshape((nr, bytes_per_row))
    style = style.reshape((nr, bytes_per_row))
    array = np.empty((nr, bytes_per_row, 3), dtype=np.uint8)
    array[:,:] = segment_viewer.preferences.empty_background_color.Get(False)
    selected_color = segment_viewer.preferences.highlight_background_color

    current_frame_number = segment_viewer.current_frame_number
    print(f"SLOW VERSION OF get_numpy_memory_access_image; frame = {current_frame_number}")
    print(source)

    from ..debugger import dtypes as dd

    for j in range(nr):
        for i in range(bytes_per_row):
            c = source[j, i]
            s = style[j, i]

            rgb = ((np.asarray(dd.default_access_type_colors[s])*c)/256).astype(np.uint8)
            if s & style_bits.selected_bit_mask:
                rgb = rgb >> 8
            array[j,i,:] = rgb
    return array


bitmap_renderer_list = [
    OneBitPerPixelB(),
    OneBitPerPixelW(),
    OneBitPerPixelPM1(),
    OneBitPerPixelPM2(),
    OneBitPerPixelPM4(),
    OneBitPerPixelApple2Linear(),
    ModeB(),
    ModeC(),
    ModeD(),
    ModeE(),
    GTIA9(),
    GTIA10(),
    GTIA11(),
    TwoBitsPerPixel(),
    FourBitsPerPixel(),
    TwoBitPlanesLE(),
    TwoBitPlanesLineLE(),
    TwoBitPlanesBE(),
    TwoBitPlanesLineBE(),
    ThreeBitPlanesLE(),
    ThreeBitPlanesLineLE(),
    ThreeBitPlanesBE(),
    ThreeBitPlanesLineBE(),
    FourBitPlanesLE(),
    FourBitPlanesLineLE(),
    FourBitPlanesBE(),
    FourBitPlanesLineBE(),
]

valid_bitmap_renderers = {item.name: item for item in bitmap_renderer_list}



# SHMEM input template locations
import numpy as np

VIDEO_WIDTH = 336
VIDEO_HEIGHT = 240

VIDEO_SIZE = VIDEO_WIDTH * VIDEO_HEIGHT
AUDIO_SIZE = 2048
MAIN_MEMORY_SIZE = 1<<16
STATESAV_MAX_SIZE = 210000

OUTPUT_DTYPE = np.dtype([
    ("video", np.uint8, VIDEO_SIZE),
    ("audio", np.uint8, AUDIO_SIZE),
    ("tag_size", np.uint32),
    ("tag_cpu", np.uint32),
    ("tag_pc", np.uint32),
    ("tag_base_ram", np.uint32),
    ("tag_base_ram_attrib", np.uint32),
    ("tag_antic", np.uint32),
    ("tag_gtia", np.uint32),
    ("tag_pia", np.uint32),
    ("tag_pokey", np.uint32),
    ("tag_filler", np.uint32, 32 - 9),
    ("flag_selftest_enabled", np.uint8),
    ("flag_filler", np.uint8, 127),
    ("state", np.uint8, STATESAV_MAX_SIZE),
])

CPU_DTYPE = np.dtype([
    ("A", np.uint8),
    ("X", np.uint8),
    ("Y", np.uint8),
    ("SP", np.uint8),
    ("P", np.uint8),
    ("PC", '<u2'),
    ])

ANTIC_DTYPE = np.dtype([
    ("DMACTL", np.uint8),
    ("CHACTL", np.uint8),
    ("HSCROL", np.uint8),
    ("VSCROL", np.uint8),
    ("PMBASE", np.uint8),
    ("CHBASE", np.uint8),
    ("NMIEN", np.uint8),
    ("NMIST", np.uint8),
    ("_ir", np.uint8),
    ("_anticmode", np.uint8),
    ("_dctr", np.uint8),
    ("_lastline", np.uint8),
    ("_need_dl", np.uint8),
    ("_vscrol_off", np.uint8),
    ("DLIST", np.uint16),
    ("_screenaddr", np.uint16),
    ("xpos", np.uint32),
    ("_xpos_limit", np.uint32),
    ("ypos", np.uint32),
    ])

GTIA_DTYPE = np.dtype([
    ("HPOSP0", np.uint8),
    ("HPOSP1", np.uint8),
    ("HPOSP2", np.uint8),
    ("HPOSP3", np.uint8),
    ("HPOSM0", np.uint8),
    ("HPOSM1", np.uint8),
    ("HPOSM2", np.uint8),
    ("HPOSM3", np.uint8),
    ("PF0PM", np.uint8),
    ("PF1PM", np.uint8),
    ("PF2PM", np.uint8),
    ("PF3PM", np.uint8),
    ("M0PL", np.uint8),
    ("M1PL", np.uint8),
    ("M2PL", np.uint8),
    ("M3PL", np.uint8),
    ("P0PL", np.uint8),
    ("P1PL", np.uint8),
    ("P2PL", np.uint8),
    ("P3PL", np.uint8),
    ("SIZEP0", np.uint8),
    ("SIZEP1", np.uint8),
    ("SIZEP2", np.uint8),
    ("SIZEP3", np.uint8),
    ("SIZEM", np.uint8),
    ("GRAFP0", np.uint8),
    ("GRAFP1", np.uint8),
    ("GRAFP2", np.uint8),
    ("GRAFP3", np.uint8),
    ("GRAFM", np.uint8),
    ("COLPM0", np.uint8),
    ("COLPM1", np.uint8),
    ("COLPM2", np.uint8),
    ("COLPM3", np.uint8),
    ("COLPF0", np.uint8),
    ("COLPF1", np.uint8),
    ("COLPF2", np.uint8),
    ("COLPF3", np.uint8),
    ("COLBK", np.uint8),
    ("PRIOR", np.uint8),
    ("VDELAY", np.uint8),
    ("GRACTL", np.uint8),
    ])

PIA_DTYPE = np.dtype([
    ("PACTL", np.uint8),
    ("PBCTL", np.uint8),
    ("PORTA", np.uint8),
    ("PORTB", np.uint8),
    ])

POKEY_DTYPE = np.dtype([
    ("KBCODE", np.uint8),
    ("IRQST", np.uint8),
    ("IRQEN", np.uint8),
    ("SKCTL", np.uint8),
    ("_shift_key", np.uint32),
    ("_keypressed", np.uint32),
    ("_DELAYED_SERIN_IRQ", np.uint32),
    ("_DELAYED_SEROUT_IRQ", np.uint32),
    ("_DELAYED_XMTDONE_IRQ", np.uint32),
    ("AUDF1", np.uint8),
    ("AUDF2", np.uint8),
    ("AUDF3", np.uint8),
    ("AUDF4", np.uint8),
    ("AUDC1", np.uint8),
    ("AUDC2", np.uint8),
    ("AUDC3", np.uint8),
    ("AUDC4", np.uint8),
    ("AUDCTL", np.uint8),
    ("_DivNIRQ", np.uint32),
    ("_DivNMax", np.uint32),
    ("_Base_mult", np.uint32),
    ])



import os
import hashlib
import inspect
import pkg_resources

import numpy as np

from . import errors
from . import style_bits
from . import utils
from .segment import Segment
from . import media_type
from . import filesystem
from .compressor import guess_compressor_list, compress_in_reverse_order, Uncompressed
from .filesystem import Dirent

import logging
log = logging.getLogger(__name__)



class ContainerHeader(Segment):
    format = np.dtype([])
    file_format = "No"

    def __init__(self, container):
        self.sector_size = 0
        self.image_size = 0
        Segment.__init__(self, container, 0, name=f"{self.file_format} Header", length=self.header_length)
        self.decode()

    @property
    def header_length(self):
        return self.format.itemsize

    def decode(self):
        size = self.header_length
        header = self.container[0:size]
        if len(header) == size:
            self.decode_from_bytes(header)
        else:
            raise errors.InvalidHeader(f"incorrect header size {len(header)} for {self.container_format}; should be {size}")

    def decode_from_bytes(self, raw):
        """Parse raw bytes and populate object attributes
        """
        raise NotImplementedError(f"decode_from_bytes not implemented for {self.file_format}")

    def encode_to_bytes(self, raw):
        """Convert values stored in header object into raw bytes in the
        container
        """
        raise NotImplementedError(f"encode_to_bytes not implemented for {self.file_format}")

    def check_media(self, media):
        if self.sector_size > 0 and self.sector_size != media.sector_size:
            raise errors.InvalidHeader("ExpectedMismatch between sector sizes: header claims {self.sector_size}, expected {media.sector_size} for {media.ui_name}")
        media_size = len(media) - self.header_length
        if self.image_size > 0 and self.image_size != media_size:
            raise errors.InvalidHeader("Invalid media size: header claims {self.image_size}, expected {media_size} for {media.ui_name}")


class Container:
    """Media storage container and packer/unpacker for disk image compression.

    Instances of this class hold a contiguous block data that represent the
    disk, cassette or cartridge image. Views of this data are in the form of
    `Segment`s which only refer to this data via a mapping of indexes into this
    container. Segments do not hold copies of the data. All operations on
    segments actually affect the container's data, and because all segments
    point to the container's data, a change to one segment can affect many
    other segments.

    In their native data format, disk images may be stored as raw data or can
    be compressed by any number of techniques. Subclasses of `Compressor`
    are used to transform compressed data into uncompressed bytes.

    Compressors may be chained, so the output of one compressor may become the
    input to the next. The order will be reversed upon saving, for example: a
    `dcm.gz` image will first be decompressed with the `gzip` algorithm, and
    that output will be in turn decompressed with the `DCM` algorithm. When
    writing the disk image back out to a file, it will be compressed with `DCM`
    before being `gzip`ped.
    """
    ui_name = "Raw Data"

    def __init__(self, data, decompression_order=None, style=None, origin=0, name="D1", error=None, verbose_name=None, memory_map=None, disasm_type=None, default_disasm_type=None, force_numpy_data=False):

        self.init_empty()
        if default_disasm_type is not None:
            self.default_disasm_type = default_disasm_type
        if force_numpy_data:
            # don't copy the data; use the data passed in as the raw data for
            # this container
            self._data = data
        else:
            self.data = data
        self.style = style
        self.disasm_type = disasm_type
        if decompression_order is None:
            decompression_order = [Uncompressed()]
        self.decompression_order = decompression_order

        self.origin = int(origin)  # force python int to decouple from possibly being a numpy datatype
        self.error = error
        self.name = name
        self.verbose_name = verbose_name
        if memory_map is not None:
            self.memory_map = memory_map

    def init_empty(self):
        self.segments = []
        self.header = None
        self._filesystem = None
        self._media = None
        self.mime = "application/octet-stream"
        self.pathname = ""
        self.default_disasm_type = 128  # libudis flag to use default CPU
        self.origin = 0
        self.error = ""
        self.name = ""
        self.verbose_name = ""
        self.memory_map = {}
        self.comments = {}
        self.uuid = utils.uuid()

        self._data = None
        self._style = None
        self._disasm_type = None

    #### properties

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, unpacked):
        if self._data is not None:
            raise errors.ReadOnlyContainer("container already populated with data")
        self._data = utils.to_numpy(unpacked)

    @property
    def style(self):
        return self._style

    @style.setter
    def style(self, value):
        if value is None:
            value = np.zeros(len(self._data), dtype=np.uint8)
        self._style = utils.to_numpy(value)

    @property
    def disasm_type(self):
        return self._disasm_type

    @disasm_type.setter
    def disasm_type(self, value):
        if value is None:
            value = np.zeros(len(self._data), dtype=np.uint8) + self.default_disasm_type
        self._disasm_type = utils.to_numpy(value)

    @property
    def sha1(self):
        return hashlib.sha1(self.data).digest()

    @property
    def header_length(self):
        return len(self.header) if self.header is not None else 0

    @property
    def media(self):
        return self._media

    @media.setter
    def media(self, value):
        self._media = value
        self._filesystem = None

    @property
    def filesystem(self):
        return self._filesystem

    @filesystem.setter
    def filesystem(self, value):
        self._filesystem = value
        self.segments = []
        if self.header:
            self.segments.append(self.header)
        if self.media:
            self.segments.append(self.media)

    @property
    def verbose_info(self):
        return self.container_info()

    def container_info(self, indent=""):
        lines = []
        name = self.verbose_name or self.name
        desc = f"{indent}{name}: {self.basename}, {len(self)} bytes, compression={','.join([c.compression_algorithm for c in self.decompression_order])}"
        lines.append(desc)
        for s in self.segments:
            v = s.segment_info(indent + "    ")
            lines.extend(v)
        return "\n".join(lines)

    @property
    def basename(self):
        return os.path.basename(self.pathname)

    #### dunder methods

    def __str__(self):
        return f"{self.basename}, size={len(self)}"

    def __len__(self):
        return np.alen(self._data)

    def __and__(self, other):
        return self._data & other

    def __iand__(self, other):
        self._data &= other
        return self

    def __getitem__(self, index):
        return self._data[index]

    def __setitem__(self, index, value):
        self._data[index] = value

    #### iterators

    def iter_segments(self):
        for segment in self.segments:
            yield segment
            yield from segment.iter_segments()

    def iter_menu(self, level):
        for segment in self.segments:
            yield (segment, level)
            yield from segment.iter_menu(level + 1)

    def iter_dirents(self):
        for segment in self.media.segments:
            if isinstance(segment, Dirent):
                yield segment
            yield from segment.iter_segments(Dirent)

    #### compression

    def calc_unpacked_bytes(self, byte_data):
        """Attempt to unpack `byte_data` using this unpacking algorithm.

        `byte_data` is a byte string, and should return a byte string if
        successfully unpacked. Conversion to a numpy array will take place
        automatically, outside of this method.

        If the data is not recognized by this subclass, raise an
        InvalidContainer exception. This signals to the caller that a different
        container type should be tried.

        If the data is recognized by this subclass but the unpacking algorithm
        is not implemented, raise an UnsupportedContainer exception. This is
        different than the InvalidContainer exception because it indicates that
        the data was indeed recognized by this subclass (despite not being
        unpacked) and checking further containers is not necessary.
        """
        return byte_data

    def calc_packed_bytes(self, skip_missing_compressors=False):
        """Pack this container into a compressed data array using this packing
        algorithm.

        Can raise `InvalidAlgorithm` if one of the compressors is read-only
        (i.e. can only decompress data). However, if `skip_missing_compressors`
        is True, no error will be raised and compression will take place
        ignoring any compressors that can't compress data.
        """
        byte_data = self.data.tobytes()
        return compress_in_reverse_order(byte_data, self.decompression_order, self.media, skip_missing_compressors)

    #### media

    def guess_media_type(self):
        media = media_type.guess_media_type(self)
        self.media = media

    def guess_filesystem(self):
        fs = filesystem.guess_filesystem(self.media)
        self.filesystem = fs

    #### serialization

    def __getstate__(self):
        """Custom jsonpickle state save routine

        This routine culls down the list of attributes that should be
        serialized, and in some cases changes their format slightly so they
        have a better mapping to json objects. For instance, json can't handle
        dicts with integer keys, so dicts are turned into lists of lists.
        Tuples are also turned into lists because tuples don't have a direct
        representation in json, while lists have a compact representation in
        json.
        """
        state = dict()
        state['__size__'] = len(self)
        state['origin'] = int(self.origin)
        state['name'] = self.name
        state['verbose_name'] = self.verbose_name
        state['uuid'] = self.uuid
        state['error'] = self.error
        state['decompression_order'] = self.decompression_order
        state['default_disasm_type'] = int(self.default_disasm_type)
        state['memory_map'] = sorted([list(i) for i in self.memory_map.items()])

        # json serialization doesn't allow int keys, so convert to list of
        # pairs
        state['comments'] = self.get_sorted_comments()

        state['disasm_type'] = utils.collapse_values(self._disasm_type)

        state['header'] = self.header
        state['media'] = self.media
        state['filesystem'] = self.filesystem
        return state

    def __setstate__(self, state):
        """Custom jsonpickle state restore routine

        The use of jsonpickle to recreate objects doesn't go through __init__,
        so there will be missing attributes when restoring old versions of the
        json. Once a version gets out in the wild and additional attributes are
        added to a segment, a default value should be applied here.
        """
        self.init_empty()
        size = state.pop('__size__')
        raw = np.zeros(size, dtype=np.uint8)
        self._data = raw
        self.style = None
        self.default_disasm_type = state.pop('default_disasm_type')
        self.disasm_type = None
        self.origin = state.pop('origin', 0)
        self.name = state.pop('name', "")
        self.verbose_name = state.pop('verbose_name', "")
        self.uuid = state.pop('uuid', utils.uuid())
        self.error = state.pop('error', "")
        self.decompression_order = state.pop('decompression_order', [])
        self.memory_map = dict(state.pop('memory_map', []))
        self.restore_comments(state.pop('comments', []))
        utils.restore_values(self._disasm_type, state.pop('disasm_type', []))

        self.header = state.pop('header')
        if self.header is not None:
            self.header.container = self
        self.restore_media(state)
        self.restore_filesystem(state)
        self.restore_containers()

        self.restore_backward_compatible_state(state)
        self.restore_missing_state()
        self.__dict__.update(state)
        self.restore_renamed_attributes()
        self.update_data_style_from_disasm_type()

    def restore_media(self, state):
        self.media = state.pop('media')

    def restore_filesystem(self, state):
        self.filesystem = state.pop('filesystem')

    def restore_containers(self):
        if self.media:
            self.media.container = self
        for segment in self.iter_segments():
            segment.container = self
            segment.restore_computed_defaults()

    def restore_backward_compatible_state(self, state):
        # convert old atrcopy stuff
        d = state.get('data ranges', None)
        if d is not None:
            utils.restore_value_to_ranges(self._disasm_type, d, 0)
        d = state.get('display list ranges', None)
        if d is not None:
            utils.restore_value_to_ranges(self._disasm_type, d, 30)

        d2 = state.get('user style 2', None)  # display list
        if d2 is not None:
            utils.restore_value_to_ranges(self._disasm_type, d2, 30)
        d3 = state.get('user style 3', None)  # jumpman level
        if d3 is not None:
            utils.restore_value_to_ranges(self._disasm_type, d3, 31)

        if d2 is not None and d2 == d3:
            # it seems that user styles 2 and 3 can point to the same ranges,
            # although 2 was supposed to be the display list and 3 the jumpman
            # level data. Try to disambiguate them because display lists
            # usually start out with 0x70, 0x70.
            for start, end in d2:
                if self._data[start] == 0x70 and self._data[start + 1] == 0x70:
                    self._disasm_type[start:end] = 30
                elif self._data[start] in [0xfc, 0xfd, 0xfe]:
                    self._disasm_type[start:end] = 32
                else:
                    self._disasm_type[start:end] = 30

        d = state.get('user style 4', None)  # jumpman harvest
        if d is not None:
            utils.restore_value_to_ranges(self._disasm_type, d, 31)

        d = state.get('comments', None)  # jumpman harvest
        if d is not None:
            self.restore_comments(d, overwrite=False)

        self.update_data_style_from_disasm_type()

    def restore_missing_state(self):
        """Hook for the future when extra serializable attributes are added to
        subclasses so new versions of the code can restore old saved files by
        providing defaults to any missing attributes.
        """
        pass

    def restore_renamed_attributes(self):
        """Hook for the future if attributes have been renamed. The old
        attribute names will have been restored in the __dict__.update in
        __setstate__, so this routine should move attribute values to their new
        names.
        """
        if hasattr(self, 'start_addr'):
            self.origin = self.start_addr
            log.debug(f"moving start_addr to origin: {self.start_addr}")
            delattr(self, 'start_addr')

    #### style

    def set_style_at_indexes(self, indexes, **kwargs):
        bits = style_bits.get_style_bits(**kwargs)
        self._style[indexes] |= bits

    def clear_style_at_indexes(self, indexes, **kwargs):
        style_mask = style_bits.get_style_mask(**kwargs)
        self.style[indexes] &= style_mask

    def get_style_ranges(self, **kwargs):
        """Return a list of start, end pairs that match the specified style
        """
        bits = style_bits.get_style_bits(**kwargs)
        matches = (self.style & bits) == bits
        return utils.bool_to_ranges(matches)

    def update_data_style_from_disasm_type(self):
        mask = style_bits.get_style_mask(data=True)
        self._style &= mask
        bits = style_bits.get_style_bits(data=True)
        indexes = np.where((self._disasm_type == 0) | ((self._disasm_type >= 30) & (self._disasm_type < 128)))[0]
        self._style[indexes] |= bits


    #### comments

    def clear_comments(self, indexes):
        mask = style_bits.get_style_mask(comment=True)
        style = self.style
        subset = style[indexes]
        comment_subset_indexes = np.where(subset & style_bits.comment_bit_mask)[0]
        print(f"clear_comments: deleting comments at {indexes}")
        style[indexes] &= mask
        for subset_index in comment_subset_indexes:
            del self.comments[indexes[subset_index]]

    def get_sorted_comments(self):
        return sorted([[int(k), str(v)] for k, v in self.comments.items()])

    def restore_comments(self, comments_list, overwrite=True):
        if overwrite:
            self.comments = {}
        bits = style_bits.get_style_bits(comment=True)
        style = self.style
        for k, v in comments_list:
            self.comments[k] = v
            style[k] |= bits

    def fixup_comments(self):
        """Remove any style bytes that are marked as commented but have no
        comment, and add any style bytes where there's a comment but it isn't
        marked in the style data.

        This happens on the base data, so only need to do this on one segment
        that uses this base data.
        """
        style_base = self._style
        comment_text_indexes = np.asarray(list(self.comments.keys()), dtype=np.uint32)
        comment_mask = style_bits.get_style_mask(comment=True)
        has_comments = np.where(style_base & style_bits.comment_bit_mask > 0)[0]
        both = np.intersect1d(comment_text_indexes, has_comments)
        log.info("fixup comments: %d correctly marked, %d without style, %d empty text" % (np.alen(both), np.alen(comment_text_indexes) - np.alen(both), np.alen(has_comments) - np.alen(both)))
        style_base &= comment_mask
        comment_style = style_bits.get_style_bits(comment=True)
        style_base[comment_text_indexes] |= comment_style

    #### search utilities

    def find_dirent(self, filename, match_case=False):
        for dirent in self.iter_dirents():
            dirent_name = dirent.filename
            if not match_case:
                dirent_name = dirent_name.lower()
                filename = filename.lower()
            print(f"checking {dirent_name} for {filename}")
            if dirent_name == filename:
                return dirent
        return None

    def find_uuid(self, uuid):
        for segment in self.iter_segments():
            if uuid == segment.uuid:
                return segment
        else:
            raise errors.InvalidSegment(f"No segment with uuid={uuid}")

def guess_container(data):
    if len(data) == 0:
        raise IOError("No data")
    data, compressors = guess_compressor_list(data)
    container = Container(data, compressors)
    return container


def load(pathname):
    sample_data = np.fromfile(pathname, dtype=np.uint8)
    container = guess_container(sample_data)
    container.pathname = pathname
    container.guess_media_type()
    return container



import pkg_resources

import numpy as np

import logging
log = logging.getLogger(__name__)


_signatures = None

def _find_signatures():
    signatures = []
    for entry_point in pkg_resources.iter_entry_points('atrip.signatures'):
        mod = entry_point.load()
        log.debug(f"find_signatures: Found module {entry_point.name}={mod.__name__}")
        if hasattr(mod, "sha1_signatures"):
            signatures.append(mod)
    return signatures

def find_signatures():
    global _signatures

    if _signatures is None:
        _signatures = _find_signatures()
    return _signatures


class Signature:
    def __init__(self, mime, name):
        self.mime = mime
        self.name = name

    def __str__(self):
        return f"{self.mime}: {self.name}"


def guess_signature_from_container(container, verbose=False):
    found = None
    sha_hash = container.sha1
    log.debug(f"container: {container}; sha1={sha_hash}")
    for mod in find_signatures():
        for mime, sigs in mod.sha1_signatures.items():
            try:
                name = sigs[sha_hash]
            except KeyError:
                continue
            else:
                log.debug(f"found match: {name}")
                return Signature(mime, name)
    else:
        log.debug(f"no match found in sha1 signature database")
    return None


# different than the above mime_parse_order, this list is the order in which
# the mime parsers will appear in a UI. Some, like the vectrex and atari2600
# parsers, aren't included in the parse order because they will identify
# many things incorrectly. They are used only when parsing by size and
# signature.
mime_display_order = [
    "application/x.atari8bit.atr",
    "application/x.atari8bit.xex",
    "application/x.atari8bit.cart",
    "application/x.atari5200.cart",
    "application/x.atari2600.cart",
    "application/x.atari2600.starpath",
    "application/x.vectrex",
    "application/x.mame_rom",
    "application/x.apple2.dsk",
    "application/x.apple2.bin",
    "application/x.rom",
    ]

pretty_mime = {
    "application/x.atari8bit.atr": "Atari 8-bit Disk Image",
    "application/x.atari8bit.xex": "Atari 8-bit Executable",
    "application/x.atari8bit.cart": "Atari 8-bit Cartridge",
    "application/x.atari5200.cart": "Atari 5200 Cartridge",
    "application/x.atari2600.cart": "Atari 2600 Cartridge",
    "application/x.atari2600.starpath": "Atari 2600 Starpath Cassette",
    "application/x.vectrex": "GCE Vectrex Cartridge",
    "application/x.mame_rom": "MAME",
    "application/x.rom": "ROM Image",
    "application/x.apple2.dsk": "Apple ][ Disk Image",
    "application/x.apple2.bin": "Apple ][ Binary",
}



import jsonpickle
import jsonpickle.ext.numpy as jsonpickle_numpy
jsonpickle_numpy.register_handlers()

import logging
log = logging.getLogger(__name__)


class Serializable:
    name = "<name>"

    # list of attributes needed to serialize the object
    serializable_attributes = []

    # set of attributes that can't be restored by simply setting the attribute
    serializable_computed = set()

    def __init__(self):
        self.init_computed_attributes(self)

    def serialize_to_dict(self):
        """Custom jsonpickle state save routine

        This routine culls down the list of attributes that should be
        serialized, and in some cases changes their format slightly so they
        have a better mapping to json objects. For instance, json can't handle
        dicts with integer keys, so dicts are turned into lists of lists.
        Tuples are also turned into lists because tuples don't have a direct
        representation in json, while lists have a compact representation in
        json.
        """
        state = dict()
        if self.name is not None:
            state['name'] = self.name
        for kls in self.__class__.__mro__:
            if hasattr(kls, 'serializable_attributes'):
                for key in kls.serializable_attributes:
                    v = getattr(self, key)
                    if key in kls.serializable_computed:
                        v = self.calc_computed_attribute(key)
                    else:
                        v = getattr(self, key)
                    state[key] = v
        return state

    def serialize_to_text(self):
        return jsonpickle.encode(self.serialize_to_dict())

    def calc_computed_attribute(self, key):
        return getattr(self, key).copy()

    def restore_from_dict(self, state):
        """Custom jsonpickle state restore routine

        The use of jsonpickle to recreate objects doesn't go through __init__,
        so there will be missing attributes when restoring old versions of the
        json. Once a version gets out in the wild and additional attributes are
        added to a segment, a default value should be applied here.
        """
        if self.name is not None and state['name'] != self.name:
            raise TypeError(f"Can't restore {state['name']} to {self.name}")
        already_seen = set()
        missing = []
        for kls in self.__class__.__mro__:
            if hasattr(kls, 'serializable_attributes'):
                print(f"restoring {kls}: {kls.serializable_attributes}")
                for key in kls.serializable_attributes:
                    if key in already_seen or key:
                        print(f"already restored {key}")
                        continue
                    elif key in kls.serializable_computed:
                        print(f"will compute value for {key} at the end")
                        continue
                    try:
                        setattr(self, key, state[key])
                        print(f"restored {key}: {getattr(self, key)}")
                        already_seen.add(key)
                    except KeyError:
                        print(f"missing key: {key}")
                        missing.append(key)
        self.restore_missing_attributes(state, missing)
        self.restore_renamed_attributes()
        self.restore_computed_attributes(state)

    def restore_from_text(self, text):
        state = jsonpickle.decode(text)
        self.restore_from_dict(state)

    def restore_computed_attributes(self, state):
        """Hook to initialize/restore attributes that depend on serialized
        attributes but aren't themselves serialized. Called during __init__ and
        after all serialized attributes have been restored to this subclass.
        """
        pass

    def restore_missing_attributes(self, state, missing):
        """Hook for the future when extra serializable attributes are added to
        subclasses so new versions of the code can restore old saved files by
        providing defaults to any missing attributes.
        """
        pass

    def restore_renamed_attributes(self):
        """Hook for the future if attributes have been renamed. The old
        attribute names will have been restored in the __dict__.update in
        __setstate__, so this routine should move attribute values to their new
        names.
        """
        pass
